#include  "evaluator.h"
#include  <cstdlib>
#include  <stdexcept>
#include  <cctype>
#include <sstream>
#include <iomanip>
#include  "intutils.h"
#include  "smithnormal.h"

namespace atlas { namespace interpreter {

const type_expr rat_type(rational_type);
const type_expr str_type(string_type);
const type_expr vec_type(vector_type);
const type_expr ratvec_type(rational_vector_type);
const type_expr mat_type(matrix_type);
const type_expr row_of_int_type(copy(int_type));
const type_expr row_of_rat_type(copy(rat_type));
const type_expr row_of_vec_type(copy(vec_type));
const type_expr row_row_of_int_type(copy(row_of_int_type));
const type_expr pair_type(*unknown_tuple(2));  // copy and destroy original
const type_expr int_int_type(*make_type("(int,int)")); 

Id_table* global_id_table=NULL;

overload_table* global_overload_table=NULL;

std::ostream* output_stream= &std::cout;

int verbosity=0;

namespace {

bindings* id_context;

context_ptr execution_context;


Hash_table::id_type size_of_name()
{ static Hash_table::id_type name=main_hash_table->match_literal("#");
  return name;
}
Hash_table::id_type print_name()
{ static Hash_table::id_type name=main_hash_table->match_literal("print");
  return name;
}
Hash_table::id_type prints_name()
{ static Hash_table::id_type name=main_hash_table->match_literal("prints");
  return name;
}

inline bool is_special_operator(Hash_table::id_type id)
{ return id==size_of_name()
        or id==print_name()
        or id==prints_name(); }

bool can_coerce_arg
  (expression e,size_t i,const type_expr& from,const type_expr& to)
{ tuple_expression* tup= dynamic_cast<tuple_expression*>(e);
  if (tup==NULL) return false;
  expression_ptr comp(tup->component[i]); tup->component[i]=NULL;
  bool result = coerce(from,to,comp); tup->component[i]=comp.release();
  return result;
}

id_pat copy_id_pat(const id_pat& p)
{
  id_pat result=p; // shallow copy
  try
  { if ((p.kind&0x2)!=0)
    { result.sublist=NULL;
      for (patlist s=p.sublist,*d=&result.sublist;
           s!=NULL; s=s->next,d=&(*d)->next)
      {
        *d=new pattern_node; (*d)->next=NULL; (*d)->body=copy_id_pat(s->body);
      }
    }
  }
  catch (...) { destroy_id_pat(&result); throw; }
  return result;
}

type_list_ptr pattern_list(const patlist p);
type_ptr pattern_type(const id_pat& pat)
{ return (pat.kind&0x2)==0
  ? copy(unknown_type)
  : make_tuple_type(pattern_list(pat.sublist));
}

type_list_ptr pattern_list(const patlist p)
{ return p==NULL ? type_list_ptr(NULL)
  : make_type_list(pattern_type(p->body),pattern_list(p->next));
}

size_t count_identifiers(const id_pat& pat)
{ size_t result= pat.kind & 0x1; // 1 if |pat.name| is defined, 0 otherwise
  if ((pat.kind & 0x2)!=0) // then a list of subpatterns is present
    for (patlist p=pat.sublist; p!=NULL; p=p->next)
      result+=count_identifiers(p->body);
  return result;
}

void list_identifiers(const id_pat& pat, std::vector<Hash_table::id_type>& d)
{ if ((pat.kind & 0x1)!=0)
    d.push_back(pat.name);
  if ((pat.kind & 0x2)!=0) // then a list of subpatterns is present
    for (patlist p=pat.sublist; p!=NULL; p=p->next)
      list_identifiers(p->body,d);
}

void thread_bindings
(const id_pat& pat,const type_expr& type, bindings& dst)
{ if ((pat.kind & 0x1)!=0) dst.add(pat.name,copy(type));
  if ((pat.kind & 0x2)!=0)
  { assert(type.kind==tuple_type);
    type_list l=type.tuple;
    for (patlist p=pat.sublist; p!=NULL; p=p->next,l=l->next)
      thread_bindings(p->body,l->t,dst);
  }
}

void thread_components
(const id_pat& pat,const shared_value& val, std::vector<shared_value>& dst)
{ if ((pat.kind & 0x1)!=0) dst.push_back(val);
  if ((pat.kind & 0x2)!=0)
  { tuple_value* t=force<tuple_value>(val.get());
    size_t i=0;
    for (patlist p=pat.sublist; p!=NULL; p=p->next,++i)
      thread_components(p->body,t->val[i],dst);
  }
}

inline bool spec_func(type_expr& t, const type_expr& from, const type_expr& to)
{ return t.specialise(gen_func_type) 
  and t.func->arg_type.specialise(from) 
  and t.func->result_type.specialise(to);
}

matrix::Vector<int> row_to_weight(const row_value& r)
{ matrix::Vector<int> result(r.val.size());
  for(size_t i=0; i<r.val.size(); ++i)
    result[i]=force<int_value>(r.val[i].get())->val;
  return result;
}

void vector_convert()
{ shared_row r(get<row_value>());
  push_value(new vector_value(row_to_weight(*r)));
}

void matrix_convert()
{ shared_row r(get<row_value>());
std::vector<matrix::Vector<int> > column_list;
  column_list.reserve(r->val.size());
  size_t depth=0; // maximal length of vectors
  for(size_t i=0; i<r->val.size(); ++i)
  { column_list.push_back(force<vector_value>(r->val[i].get())->val);
    if (column_list[i].size()>depth) depth=column_list[i].size();
  }
  for(size_t i=0; i<column_list.size(); ++i)
    // extend weights with 0's if necessary
    if (column_list[i].size()<depth)
    { size_t j=column_list[i].size();
      column_list[i].resize(depth);
      for (;j<depth; ++j) column_list[i][j]=0;
    }
  push_value(new matrix_value(matrix::Matrix<int>(column_list)));
}

void rational_convert()
{ shared_int i = get<int_value>();
    push_value(new rat_value(arithmetic::Rational(i->val)));
}

void ratvec_convert()
{ shared_row r = get <row_value>();
  matrix::Vector<int> numer(r->val.size()),denom(r->val.size());
  unsigned int d=1;
  for (size_t i=0; i<r->val.size(); ++i)
  { arithmetic::Rational frac = force<rat_value>(r->val[i].get())->val;
    numer[i]=frac.numerator();
    denom[i]=frac.denominator();
    d=arithmetic::lcm(d,denom[i]);
  }
  for (size_t i=0; i<r->val.size(); ++i)
    numer[i]*= d/denom[i];

  push_value(new rational_vector_value(latticetypes::RatWeight(numer,d)));
}

void rat_list_convert()
{ shared_rational_vector rv = get<rational_vector_value>();
  row_ptr result(new row_value(rv->val.size()));
  for (size_t i=0; i<rv->val.size(); ++i)
  { arithmetic::Rational q(rv->val.numerator()[i],rv->val.denominator());
    result->val[i] = shared_value(new rat_value(q.normalize()));
  }
  push_value(result);
}

void matrix2_convert()
{ shared_row r(get<row_value>());
std::vector<matrix::Vector<int> > column_list;
  column_list.reserve(r->val.size());
  size_t depth=0; // maximal length of vectors
  for(size_t i=0; i<r->val.size(); ++i)
  { column_list.push_back(row_to_weight(*force<row_value>(r->val[i].get())));
    if (column_list[i].size()>depth) depth=column_list[i].size();
  }
  for(size_t i=0; i<column_list.size(); ++i)
    // extend weights with 0's if necessary
    if (column_list[i].size()<depth)
    { size_t j=column_list[i].size();
      column_list[i].resize(depth);
      for (;j<depth; ++j) column_list[i][j]=0;
    }
  push_value(new matrix_value(matrix::Matrix<int>(column_list)));

}

row_ptr weight_to_row(const matrix::Vector<int>& v)
{ row_ptr result (new row_value(v.size()));
  for(size_t i=0; i<v.size(); ++i)
    result->val[i]=shared_value(new int_value(v[i]));
  return result;
}

void int_list_convert()
{ shared_vector v(get<vector_value>());
  push_value(weight_to_row(v->val));
}

void vec_list_convert()
{ matrix::Matrix<int> m = get<matrix_value>()->val;
  row_ptr result(new row_value(m.numColumns()));
  for(size_t i=0; i<m.numColumns(); ++i)
    result->val[i]=shared_value(new vector_value(m.column(i)));
  push_value(result);
}

void int_list_list_convert()
{ matrix::Matrix<int> m = get<matrix_value>()->val;
  row_ptr result(new row_value(m.numColumns()));
  for(size_t i=0; i<m.numColumns(); ++i)
    result->val[i]=shared_value(weight_to_row(m.column(i)).release());

  push_value(result);
}

void plus_wrapper(expression_base::level l)
{ int j=get<int_value>()->val; int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
    push_value(new int_value(i+j));
}

void minus_wrapper(expression_base::level l)
{ int j=get<int_value>()->val; int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
    push_value(new int_value(i-j));
}

void times_wrapper(expression_base::level l)
{ int j=get<int_value>()->val; int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
    push_value(new int_value(i*j));
}

void divide_wrapper(expression_base::level l)
{ int j=get<int_value>()->val; int i=get<int_value>()->val;
  if (j==0) throw std::runtime_error("Division by zero");
  if (l!=expression_base::no_value)
    push_value(new int_value(i/j));
}

void modulo_wrapper(expression_base::level l)
{ int  j=get<int_value>()->val; int i=get<int_value>()->val;
  if (j==0) throw std::runtime_error("Modulo zero");
  if (l!=expression_base::no_value)
    push_value(new int_value(i%j));
}

void divmod_wrapper(expression_base::level l)
{ int j=get<int_value>()->val; int i=get<int_value>()->val;
  if (j==0) throw std::runtime_error("DivMod by zero");
  if (l!=expression_base::no_value)
  { push_value(new int_value(i/j));
    push_value(new int_value(i%j));
    if (l==expression_base::single_value)
      wrap_tuple(2);
  }
}

void unary_minus_wrapper(expression_base::level l)
{ int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
    push_value(new int_value(-i)); }

void fraction_wrapper(expression_base::level l)
{ int d=get<int_value>()->val; int n=get<int_value>()->val;
  if (d==0) throw std::runtime_error("fraction with zero denominator");
  if (l!=expression_base::no_value)
    push_value(new rat_value(arithmetic::Rational(n,d)));
}

void power_wrapper(expression_base::level l)
{ static shared_int one(new int_value(1));
int n=get<int_value>()->val; shared_int i=get<int_value>();
  if (intutils::abs(i->val)!=1 and n<0)
    throw std::runtime_error("Negative power of integer");
  if (l==expression_base::no_value)
    return;

  if (i->val==1)
  { push_value(one);
      return;
  }
  if (i->val==-1)
  { push_value(n%2==0 ? one : i);
      return;
  }

  push_value(new int_value(arithmetic::power(i->val,n)));
}

void rat_plus_wrapper(expression_base::level l)
{ arithmetic::Rational j=get<rat_value>()->val;
  arithmetic::Rational i=get<rat_value>()->val;
  if (l!=expression_base::no_value)
    push_value(new rat_value(i+j));
}

void rat_minus_wrapper(expression_base::level l)
{ arithmetic::Rational j=get<rat_value>()->val;
  arithmetic::Rational i=get<rat_value>()->val;
  if (l!=expression_base::no_value)
    push_value(new rat_value(i-j));
}

void rat_times_wrapper(expression_base::level l)
{ arithmetic::Rational j=get<rat_value>()->val;
  arithmetic::Rational i=get<rat_value>()->val;
  if (l!=expression_base::no_value)
    push_value(new rat_value(i*j));
}

void rat_divide_wrapper(expression_base::level l)
{ arithmetic::Rational j=get<rat_value>()->val;
  arithmetic::Rational i=get<rat_value>()->val;
  if (j.numerator()==0)
    throw std::runtime_error("Rational division by zero");
  if (l!=expression_base::no_value)
    push_value(new rat_value(i/j));
}

void rat_unary_minus_wrapper(expression_base::level l)
{ arithmetic::Rational i=get<rat_value>()->val;
  if (l!=expression_base::no_value)
    push_value(new rat_value(arithmetic::Rational(0)-i)); }

void rat_inverse_wrapper(expression_base::level l)
{ arithmetic::Rational i=get<rat_value>()->val;
  if (i.numerator()==0)
    throw std::runtime_error("Inverse of zero");
  if (l!=expression_base::no_value)
    push_value(new rat_value(arithmetic::Rational(1)/i)); }

void rat_power_wrapper(expression_base::level l)
{ int n=get<int_value>()->val; arithmetic::Rational b=get<rat_value>()->val;
  if (b.numerator()==0 and n<0)
    throw std::runtime_error("Negative power of zero");
  if (l!=expression_base::no_value)
    push_value(new rat_value(b.power(n)));
}

void int_eq_wrapper(expression_base::level l)
{ shared_int j=get<int_value>(); shared_int i=get<int_value>();
  if (l!=expression_base::no_value)
    push_value(new bool_value(i->val==j->val));
}

void int_neq_wrapper(expression_base::level l)
{ shared_int j=get<int_value>(); shared_int i=get<int_value>();
  if (l!=expression_base::no_value)
    push_value(new bool_value(i->val!=j->val));
}

void int_less_wrapper(expression_base::level l)
{ shared_int j=get<int_value>(); shared_int i=get<int_value>();
  if (l!=expression_base::no_value)
    push_value(new bool_value(i->val<j->val));
}

void int_lesseq_wrapper(expression_base::level l)
{ shared_int j=get<int_value>(); shared_int i=get<int_value>();
  if (l!=expression_base::no_value)
    push_value(new bool_value(i->val<=j->val));
}

void int_greater_wrapper(expression_base::level l)
{ shared_int j=get<int_value>(); shared_int i=get<int_value>();
  if (l!=expression_base::no_value)
    push_value(new bool_value(i->val>j->val));
}

void int_greatereq_wrapper(expression_base::level l)
{ shared_int j=get<int_value>(); shared_int i=get<int_value>();
  if (l!=expression_base::no_value)
    push_value(new bool_value(i->val>=j->val));
}

void rat_eq_wrapper(expression_base::level l)
{ shared_rat j=get<rat_value>(); shared_rat i=get<rat_value>();
  if (l!=expression_base::no_value)
    push_value(new bool_value(i->val==j->val));
}

void rat_neq_wrapper(expression_base::level l)
{ shared_rat j=get<rat_value>(); shared_rat i=get<rat_value>();
  if (l!=expression_base::no_value)
    push_value(new bool_value(i->val!=j->val));
}

void rat_less_wrapper(expression_base::level l)
{ shared_rat j=get<rat_value>(); shared_rat i=get<rat_value>();
  if (l!=expression_base::no_value)
    push_value(new bool_value(i->val<j->val));
}

void rat_lesseq_wrapper(expression_base::level l)
{ shared_rat j=get<rat_value>(); shared_rat i=get<rat_value>();
  if (l!=expression_base::no_value)
    push_value(new bool_value(i->val<=j->val));
}

void rat_greater_wrapper(expression_base::level l)
{ shared_rat j=get<rat_value>(); shared_rat i=get<rat_value>();
  if (l!=expression_base::no_value)
    push_value(new bool_value(i->val>j->val));
}

void rat_greatereq_wrapper(expression_base::level l)
{ shared_rat j=get<rat_value>(); shared_rat i=get<rat_value>();
  if (l!=expression_base::no_value)
    push_value(new bool_value(i->val>=j->val));
}

void equiv_wrapper(expression_base::level l)
{ bool a=get<bool_value>()->val; bool b=get<bool_value>()->val;
  if (l!=expression_base::no_value)
    push_value(new bool_value(a==b));
}

void inequiv_wrapper(expression_base::level l)
{ bool a=get<bool_value>()->val; bool b=get<bool_value>()->val;
  if (l!=expression_base::no_value)
    push_value(new bool_value(a!=b));
}

void concatenate_wrapper(expression_base::level l)
{ shared_string b=get<string_value>(); shared_string a=get<string_value>();
  if (l!=expression_base::no_value)
    push_value(new string_value(a->val+b->val));
}
void int_format_wrapper(expression_base::level l)
{ shared_int n=get<int_value>();
  std::ostringstream o; o<<n->val;
  if (l!=expression_base::no_value)
    push_value(new string_value(o.str()));
}

void print_wrapper(expression_base::level l)
{ *output_stream << *pop_value() << std::endl;
  if (l==expression_base::single_value)
    wrap_tuple(0); // don't forget to return a value if asked for
}

void prints_wrapper(expression_base::level l)
{ shared_value v=pop_value();
  string_value* s=dynamic_cast<string_value*>(v.get());
  if (s!=NULL)
    *output_stream << s->val << std::endl;
  else
  { tuple_value* t=dynamic_cast<tuple_value*>(v.get());
    if (t!=NULL)
    { for (size_t i=0; i<t->val.size(); ++i)
      { s=dynamic_cast<string_value*>(t->val[i].get());
        if (s!=NULL)
	  *output_stream << s->val;
        else
           *output_stream << *t->val[i];
      }
      *output_stream << std::endl;
    }
    else
      *output_stream << *v << std::endl; // just like |print| in other cases
  }
  if (l==expression_base::single_value)
    wrap_tuple(0); // don't forget to return a value if asked for
}

void sizeof_string_wrapper(expression_base::level l)
{ size_t s=get<string_value>()->val.size();
  if (l!=expression_base::no_value)
    push_value(new int_value(s));
}

void sizeof_vector_wrapper(expression_base::level l)
{ size_t s=get<vector_value>()->val.size();
  if (l!=expression_base::no_value)
    push_value(new int_value(s));
}


void matrix_bounds_wrapper(expression_base::level l)
{ shared_matrix m=get<matrix_value>();
  if (l==expression_base::no_value)
    return;
  push_value(new int_value(m->val.numRows()));
  push_value(new int_value(m->val.numColumns()));
  if (l==expression_base::single_value)
    wrap_tuple(2);
}

void sizeof_wrapper(expression_base::level l)
{ size_t s=get<row_value>()->val.size();
  if (l!=expression_base::no_value)
    push_value(new int_value(s));
}

void vector_suffix_wrapper(expression_base::level l)
{ int e=get<int_value>()->val;
  shared_vector r=get_own<vector_value>();
  if (l!=expression_base::no_value)
  { r->val.push_back(e);
    push_value(r);
  }
}

void vector_prefix_wrapper(expression_base::level l)
{ shared_vector r=get_own<vector_value>();
  int e=get<int_value>()->val;
  if (l!=expression_base::no_value)
  { r->val.insert(r->val.begin(),e);
    push_value(r);
  }
}

void join_vectors_wrapper(expression_base::level l)
{ shared_vector y=get<vector_value>();
  shared_vector x=get<vector_value>();
  if (l!=expression_base::no_value)
  { vector_ptr result(new vector_value(std::vector<int>()));
    result->val.reserve(x->val.size()+y->val.size());
    result->val.insert(result->val.end(),x->val.begin(),x->val.end());
    result->val.insert(result->val.end(),y->val.begin(),y->val.end());
    push_value(result);
  }

}

void suffix_element_wrapper(expression_base::level l)
{ shared_value e=pop_value();
  shared_row r=get_own<row_value>();
  if (l!=expression_base::no_value)
  { r->val.push_back(e);
    push_value(r);
  }
}

void prefix_element_wrapper(expression_base::level l)
{ shared_row r=get_own<row_value>();
  shared_value e=pop_value();
  if (l!=expression_base::no_value)
  { r->val.insert(r->val.begin(),e);
    push_value(r);
  }
}

void join_rows_wrapper(expression_base::level l)
{ shared_row y=get<row_value>();
  shared_row x=get<row_value>();
  if (l!=expression_base::no_value)
  { row_ptr result(new row_value(0));
    result->val.reserve(x->val.size()+y->val.size());
    result->val.insert(result->val.end(),x->val.begin(),x->val.end());
    result->val.insert(result->val.end(),y->val.begin(),y->val.end());
    push_value(result);
  }

}

void error_wrapper(expression_base::level l)
{ throw std::runtime_error(get<string_value>()->val); }

void diagonal_wrapper(expression_base::level l)
{ matrix::Vector<int> d=get<vector_value>()->val;
  if (l==expression_base::no_value)
    return;
  size_t n=d.size();
  matrix_ptr m (new matrix_value(matrix::Matrix<int>(n,n,0)));
  for (size_t i=0; i<n; ++i)
    m->val(i,i)=d[i];
  push_value(m);
}

void vector_div_wrapper(expression_base::level l)
{ int n=get<int_value>()->val;
  matrix::Vector<int> v=get<vector_value>()->val;
  if (l!=expression_base::no_value)
    push_value(new rational_vector_value(v,n));
}

void invfact_wrapper(expression_base::level l)
{ matrix::Matrix<int> m=get<matrix_value>()->val;
  if (l==expression_base::no_value)
    return;
  size_t nr=m.numRows();
  std::vector<matrix::Vector<int> > b;  matrix::initBasis(b,nr);
  vector_ptr inv_factors (new vector_value(matrix::Vector<int>(0)));
  smithnormal::smithNormal(inv_factors->val,b.begin(),m);
  push_value(inv_factors);
}

void Smith_basis_wrapper(expression_base::level l)
{ matrix::Matrix<int> m=get<matrix_value>()->val;
  if (l==expression_base::no_value)
    return;
  size_t nr=m.numRows();
  std::vector<matrix::Vector<int> > b;  matrix::initBasis(b,nr);
  matrix::Vector<int> inv_factors(0);
  smithnormal::smithNormal(inv_factors,b.begin(),m);
  push_value(new matrix_value(matrix::Matrix<int>(b)));
}


void Smith_wrapper(expression_base::level l)
{ matrix::Matrix<int> m=get<matrix_value>()->val;
  if (l==expression_base::no_value)
    return;
  size_t nr=m.numRows();
  std::vector<matrix::Vector<int> > b;  matrix::initBasis(b,nr);
  vector_ptr inv_factors (new vector_value(matrix::Vector<int>(0)));
  smithnormal::smithNormal(inv_factors->val,b.begin(),m);
push_value(new matrix_value(matrix::Matrix<int>(b)));
  push_value(inv_factors);
  if (l==expression_base::single_value)
    wrap_tuple(2);
}

void invert_wrapper(expression_base::level l)
{ matrix::Matrix<int> m=get<matrix_value>()->val;
  if (m.numRows()!=m.numColumns())
  { std::ostringstream s;
    s<< "Cannot invert a " 
     << m.numRows() << "x" << m.numColumns() << " matrix";
    throw std::runtime_error(s.str());
  }
  if (l==expression_base::no_value)
    return;
  int_ptr denom(new int_value(0));
push_value(new matrix_value(m.inverse(denom->val)));
  push_value(denom);
  if (l==expression_base::single_value)
    wrap_tuple(2);
}

} // |namespace|

void push_expanded(expression_base::level l, const shared_value& v)
{ if(l==expression_base::single_value)
    push_value(v);
  else if (l==expression_base::multi_value)
  { shared_tuple p = std::tr1::dynamic_pointer_cast<tuple_value>(v);
    if (p==NULL)
      push_value(v);
    else
      for (size_t i=0; i<p->length(); ++i)
        push_value(p->val[i]); // push components
  }
}

void initialise_evaluator()
{ 
  execution_stack.reserve(16); 
  
  coercion(row_of_int_type, vec_type, "V", vector_convert); 
  coercion(row_of_vec_type,mat_type, "M", matrix_convert);
  
  coercion(int_type,rat_type, "Q", rational_convert); 
  coercion(row_of_rat_type,ratvec_type, "QV", ratvec_convert); 
  coercion(ratvec_type,row_of_rat_type, "[Q]", rat_list_convert);
  
  coercion(row_row_of_int_type,mat_type, "M2", matrix2_convert); 
  coercion(vec_type,row_of_int_type, "[I]", int_list_convert); 
  coercion(mat_type,row_of_vec_type, "[V]", vec_list_convert); 
  coercion(mat_type,row_row_of_int_type, "[[I]]", int_list_list_convert); 
  
  install_function(plus_wrapper,"+","(int,int->int)");
  install_function(minus_wrapper,"-","(int,int->int)");
  install_function(times_wrapper,"*","(int,int->int)");
  install_function(divide_wrapper,"\\","(int,int->int)");
  install_function(modulo_wrapper,"%","(int,int->int)");
  install_function(divmod_wrapper,"\\%","(int,int->int,int)");
  install_function(unary_minus_wrapper,"-","(int->int)");
  install_function(fraction_wrapper,"/","(int,int->rat)");
  install_function(power_wrapper,"^","(int,int->int)");
  install_function(rat_plus_wrapper,"+","(rat,rat->rat)");
  install_function(rat_minus_wrapper,"-","(rat,rat->rat)");
  install_function(rat_times_wrapper,"*","(rat,rat->rat)");
  install_function(rat_divide_wrapper,"/","(rat,rat->rat)");
  install_function(rat_unary_minus_wrapper,"-","(rat->rat)");
  install_function(rat_inverse_wrapper,"/","(rat->rat)");
  install_function(rat_power_wrapper,"^","(rat,int->rat)");
  install_function(int_eq_wrapper,"=","(int,int->bool)");
  install_function(int_neq_wrapper,"!=","(int,int->bool)");
  install_function(int_less_wrapper,"<","(int,int->bool)");
  install_function(int_lesseq_wrapper,"<=","(int,int->bool)");
  install_function(int_greater_wrapper,">","(int,int->bool)");
  install_function(int_greatereq_wrapper,">=","(int,int->bool)");
  install_function(rat_eq_wrapper,"=","(rat,rat->bool)");
  install_function(rat_neq_wrapper,"!=","(rat,rat->bool)");
  install_function(rat_less_wrapper,"<","(rat,rat->bool)");
  install_function(rat_lesseq_wrapper,"<=","(rat,rat->bool)");
  install_function(rat_greater_wrapper,">","(rat,rat->bool)");
  install_function(rat_greatereq_wrapper,">=","(rat,rat->bool)");
  install_function(equiv_wrapper,"=","(bool,bool->bool)");
  install_function(inequiv_wrapper,"!=","(bool,bool->bool)");
  install_function(int_format_wrapper,"int_format","(int->string)");
  install_function(sizeof_string_wrapper,"#","(string->int)");
  install_function(sizeof_vector_wrapper,"#","(vec->int)");
  install_function(matrix_bounds_wrapper,"#","(mat->int,int)");
  install_function(vector_div_wrapper,"/","(vec,int->ratvec)");
  install_function(id_mat_wrapper,"id_mat","(int->mat)");
  install_function(error_wrapper,"error","(string->)");
  install_function(concatenate_wrapper,"#","(string,string->string)");
  install_function(vector_suffix_wrapper,"#","(vec,int->vec)");
  install_function(vector_prefix_wrapper,"#","(int,vec->vec)");
  install_function(join_vectors_wrapper,"#","(vec,vec->vec)");
  install_function(transpose_mat_wrapper,"^","(mat->mat)");
  install_function(transpose_vec_wrapper,"^","(vec->mat)");
  install_function(diagonal_wrapper,"diagonal_mat","(vec->mat)");
  install_function(vv_prod_wrapper,"*","(vec,vec->int)");
  install_function(mv_prod_wrapper,"*","(mat,vec->vec)");
  install_function(mm_prod_wrapper,"*","(mat,mat->mat)");
  install_function(vm_prod_wrapper,"*","(vec,mat->vec)");
  install_function(invfact_wrapper,"inv_fact","(mat->vec)");
  install_function(Smith_basis_wrapper,"Smith_basis","(mat->mat)");
  install_function(Smith_wrapper,"Smith","(mat->mat,vec)");
  install_function(invert_wrapper,"invert","(mat->mat,int)");
 }

void reset_evaluator()
{ if (!execution_stack.empty())
  { if (verbosity==0)
      execution_stack.clear();
    else
    {
      std::cerr << "Discarding from execution stack:" << std::endl;
      do
        std::cerr << *pop_value() << std::endl;
      while (!execution_stack.empty());
    }
  }
  
  id_context=NULL;
  
  execution_context.reset();
}

void denotation::evaluate(level l) const
{ if (l!=no_value) push_value(denoted_value); }

void bindings::add(Hash_table::id_type id,type_ptr t)
{ for (base::const_iterator it=begin(); it!=end(); ++it)
    if (it->first==id)
      throw program_error(std::string("Multiple binding of ")
                          +main_hash_table->name_of(id)
                          +" in same scope");
  push_back(std::make_pair(id,(type_p)NULL));
  back().second=t.release();
}

type_p bindings::lookup
  (Hash_table::id_type id, size_t& depth, size_t& offset) const
{ size_t i=0;
  for (const bindings* p=this; p!=NULL; p=p->next,++i)
    for (size_t j=0; j<p->size(); ++j)
      if ((*p)[j].first==id)
      { depth=i; offset=j; return (*p)[j].second; }
  return NULL;
}

expression convert_expr(const expr& e, type_expr& type)
  throw(std::bad_alloc,program_error)
{

  switch(e.kind)
  { case integer_denotation:
    { expression_ptr d(new denotation
        (shared_value(new int_value(e.e.int_denotation_variant))));
      if(type.specialise(int_type) or coerce(int_type,type,d))
        return d.release();
      else throw type_error(e,copy(int_type),copy(type));
    }
   case string_denotation:
    { expression_ptr d(new denotation
        (shared_value(new string_value(e.e.str_denotation_variant))));
      if (type.specialise(str_type) or coerce(str_type,type,d))
        return d.release();
      else throw type_error(e,copy(str_type),copy(type));
    }
   case boolean_denotation:
    { expression_ptr d(new denotation
          (shared_value(new bool_value(e.e.int_denotation_variant))));
      if (type.specialise(bool_type) or coerce(bool_type,type,d))
        return d.release();
      else throw type_error(e,copy(bool_type),copy(type));
    }
   
   case list_display:
     if (type.specialise(row_of_type))
     { std::auto_ptr<list_expression> result (new list_expression(0));
       result->component.reserve(length(e.e.sublist));
       for (expr_list l=e.e.sublist; l!=NULL; l=l->next)
         result->component.push_back(convert_expr(l->e,*type.component_type));
       return result.release(); // and convert (derived|->|base) to |expression|
     }
     else
     
     { type_expr comp_type;
       conversion_record* conv = row_coercion(type,comp_type);
       if (conv==NULL)
         throw type_error(e,copy(row_of_type),copy(type));
     
       std::auto_ptr<list_expression> display
           (new list_expression(length(e.e.sublist)));
       size_t i=0;
         for (expr_list l=e.e.sublist; l!=NULL; l=l->next,++i)
           display->component[i]=convert_expr(l->e,comp_type);
         return new conversion(*conv,expression_ptr(display));
     }
   
   
   case tuple_display:
   { type_ptr tup=unknown_tuple(length(e.e.sublist));
     if (type.specialise(*tup))
     { std::auto_ptr<tuple_expression> result(new tuple_expression(0));
       result->component.reserve(length(e.e.sublist));
     type_list tl=type.tuple;
       for (expr_list el=e.e.sublist; el!=NULL; el=el->next,tl=tl->next)
         result->component.push_back(convert_expr(el->e,tl->t));
       return result.release();  // and convert (derived|->|base) to |expression|
     }
     else throw type_error(e,tup,copy(type));
   }
   break;
   
   case subscription:
   { type_expr array_type, index_type, subscr_type;
       // initialised to |undetermined_type|
     expression_ptr array
       (convert_expr(e.e.subscription_variant->array,array_type));
     expression_ptr index
       (convert_expr(e.e.subscription_variant->index,index_type));
     subscr_base::sub_type kind;
     expression_ptr subscr;
     if (subscr_base::indexable(array_type,index_type,subscr_type,kind))
       switch (kind)
       { case subscr_base::row_entry:
         subscr.reset(new row_subscription(array,index));
       break;
       case subscr_base::vector_entry:
         subscr.reset(new vector_subscription(array,index));
       break;
       case subscr_base::matrix_entry:
         subscr.reset(new matrix_subscription(array,index));
       break;
       case subscr_base::matrix_column:
         subscr.reset(new matrix_slice(array,index));
       break;
       }
     else
     { std::ostringstream o;
       o << "Cannot subscript " << array_type << " value with index of type "
         << index_type;
       throw expr_error(e,o.str());
     }
   
     if (type.specialise(subscr_type) or coerce(subscr_type,type,subscr))
       return subscr.release(); // and convert (derived|->|base) to |expression|
     else throw type_error(e,copy(subscr_type),copy(type));
   }
   
   case applied_identifier:
   { type_p it; expression_ptr id; size_t i,j;
     if ((it=id_context->lookup(e.e.identifier_variant,i,j))!=NULL)
       id.reset(new local_identifier(e.e.identifier_variant,i,j));
     else if ((it=global_id_table->type_of(e.e.identifier_variant))!=NULL)
       id.reset(new global_identifier(e.e.identifier_variant));
     else throw program_error  
          (std::string("Undefined identifier ")
   	+main_hash_table->name_of(e.e.identifier_variant));
   
     if (type.specialise(*it))
       { it->specialise(type); return id.release(); }
     else if (coerce(*it,type,id))
       return id.release();
     else throw type_error(e,copy(*it),copy(type));
   }
   
   case function_call:
   { if (e.e.call_variant->fun.kind==applied_identifier)
       
       { const Hash_table::id_type id =e.e.call_variant->fun.e.identifier_variant;
         const expr arg=e.e.call_variant->arg;
         size_t i,j;
         if (not is_empty(arg) and id_context->lookup(id,i,j)==NULL)
         { const overload_table::variant_list& variants
             = global_overload_table->variants(id);
           if (variants.size()>0 or is_special_operator(id))
             return resolve_overload(e,type,variants);
         }
       }
     type_ptr f_type=copy(gen_func_type); // start with generic function type
     expression_ptr fun(convert_expr(e.e.call_variant->fun,*f_type));
     expression_ptr arg
       (convert_expr(e.e.call_variant->arg,f_type->func->arg_type));
     expression_ptr call (new call_expression(fun,arg));
     if (type.specialise(f_type->func->result_type) or
         coerce(f_type->func->result_type,type,call))
       return call.release();
     else throw type_error(e,copy(f_type->func->result_type),copy(type));
   }
   
   case let_expr:
   { let lexp=e.e.let_variant;
     id_pat& pat=lexp->pattern;
     type_ptr decl_type=pattern_type(pat);
     expression_ptr arg(convert_expr(lexp->val,*decl_type));
     size_t n_id=count_identifiers(pat);
   bindings new_bindings(n_id);
     thread_bindings(pat,*decl_type,new_bindings);
     new_bindings.push(id_context);
     expression_ptr body(convert_expr(lexp->body,type));
     new_bindings.pop(id_context);
     expression_ptr func(new lambda_expression(pat,body));
     return new call_expression(func,arg);
   }
   
   case lambda_expr:
   { if (not type.specialise(gen_func_type))
       throw type_error(e,copy(gen_func_type),copy(type));
     lambda fun=e.e.lambda_variant;
     id_pat& pat=fun->pattern;
     type_expr& arg_type=*static_cast<type_p>(fun->arg_type);
     if (not type.func->arg_type.specialise(arg_type))
     throw type_error(e,
                        make_function_type(copy(arg_type),copy(unknown_type)),
                        copy(type));
     size_t n_id=count_identifiers(pat);
   bindings new_bindings(n_id);
     thread_bindings(pat,arg_type,new_bindings);
     new_bindings.push(id_context);
     expression_ptr body(convert_expr(fun->body,type.func->result_type));
     new_bindings.pop(id_context);
     return new lambda_expression(pat,body);
   }
   
   case conditional_expr:
     { expression_ptr c (convert_expr(e.e.if_variant->condition,bool_type));
       expression_ptr el (convert_expr(e.e.if_variant->else_branch,type));
       expression_ptr th (convert_expr(e.e.if_variant->then_branch,type));
       return new conditional_expression(c,th,el);
     }
   break;
   
   case while_expr:
     { w_loop w=e.e.while_variant;
       expression_ptr c (convert_expr(w->condition,bool_type));
       expression_ptr n
        (is_empty(w->next_part) ? NULL  : convert_expr(w->next_part,void_type));
       if (type==void_type or type.specialise(row_of_type))
       { expression_ptr b
          (convert_expr(w->body, 
                        type==void_type ? void_type :*type.component_type));
         return new while_expression(c,b,n);
       }
       else
       
       { type_expr comp_type;
         conversion_record* conv = row_coercion(type,comp_type);
         if (conv==NULL)
           throw type_error(e,copy(row_of_type),copy(type));
       
         expression_ptr b(convert_expr(w->body,comp_type));
         expression_ptr loop(new while_expression(c,b,n));
         return new conversion(*conv,loop);
       }
     }
   break;
   
   case for_expr:
   { f_loop f=e.e.for_variant; type_expr in_type;
     expression_ptr in_expr(convert_expr(f->in_part,in_type));
   type_expr comp_type; subscr_base::sub_type which;
     if (not subscr_base::indexable(in_type,int_type,comp_type,which))
     { std::ostringstream o;
       o << "Cannot iterate over value of type " << in_type;
       throw expr_error(e,o.str());
     }
     type_ptr pt = pattern_type(f->id);
     type_ptr it_type=make_tuple_type(make_type_list
       (copy(int_type),make_type_singleton(copy(comp_type))));
     if (not pt->specialise(*it_type))
       throw expr_error(e,"Improper structure of loop variable pattern");
     size_t n_id = count_identifiers(f->id);
     bindings bind(n_id); thread_bindings(f->id,*it_type,bind);
     type_expr body_type, *btp; conversion_record* conv=NULL;
     if (type==void_type)
       btp=&void_type;
     else if (type.specialise(row_of_type))
       btp=type.component_type;
     else if ((conv=row_coercion(type,body_type))!=NULL)
       btp=&body_type;
     else throw type_error(e,copy(row_of_type),copy(type));
     bind.push(id_context);
     expression_ptr body(convert_expr (f->body,*btp));
     bind.pop(id_context);
     expression_ptr loop(new for_expression(f->id,in_expr,body,which));
     return conv==NULL ? loop.release() : new conversion(*conv,loop);
   }
   break;
   
   case cfor_expr:
   { c_loop c=e.e.cfor_variant;
     expression_ptr count_expr(convert_expr(c->count,int_type));
     static shared_value zero=shared_value(new int_value(0));
       // avoid repeated allocation
     expression_ptr bound_expr
       (is_empty(c->bound) ? new denotation(zero)
       : convert_expr(c->bound,int_type)
       );
   
     bindings bind(1); bind.add(c->id,copy(int_type));
     type_expr body_type, *btp; conversion_record* conv=NULL;
     if (type==void_type)
       btp=&void_type;
     else if (type.specialise(row_of_type))
       btp=type.component_type;
     else if ((conv=row_coercion(type,body_type))!=NULL)
       btp=&body_type;
     else throw type_error(e,copy(row_of_type),copy(type));
     bind.push(id_context);
     expression_ptr body(convert_expr (c->body,*btp));
     bind.pop(id_context);
     expression e;
     if (c->up!=0)
       e=new inc_for_expression(c->id,count_expr,bound_expr,body);
     else
       e=new dec_for_expression(c->id,count_expr,bound_expr,body);
     expression_ptr loop(e);
     return conv==NULL ? loop.release() : new conversion(*conv,loop);
   
   }
   break;
   
   case cast_expr:
   { cast c=e.e.cast_variant;
     type_expr& ctype=*static_cast<type_p>(c->type);
     expression_ptr p(convert_expr(c->exp,ctype));
     if (type.specialise(ctype) or coerce(ctype,type,p))
       return p.release();
     throw type_error(e,copy(ctype),copy(type));
   }
   break;
   
   case op_cast_expr:
   { op_cast c=e.e.op_cast_variant;
     const overload_table::variant_list& variants =
      global_overload_table->variants(c->oper);
     type_expr& ctype=*static_cast<type_p>(c->type);
     if (is_special_operator(c->oper))
       
       { if (c->oper==size_of_name())
         { if (ctype.kind==row_type)
           { if (spec_func(type,ctype,int_type))
             return new denotation(shared_value
               (new builtin_value(sizeof_wrapper,"#@[T]")));
             throw type_error(e,copy(ctype),copy(type));
           }
           else if (ctype.specialise(pair_type))
           { type_expr& arg_tp0 = ctype.tuple->t;
             type_expr& arg_tp1 = ctype.tuple->next->t;
             if (arg_tp0.kind==row_type)
             { if (arg_tp0==arg_tp1)
               { if (spec_func(type,ctype,arg_tp0))
                 return new denotation(shared_value 
                   (new builtin_value(join_rows_wrapper,"#@([T],[T]->[T])")));
                 throw type_error(e,copy(ctype),copy(type));
               }
       	else if (*arg_tp0.component_type==arg_tp1)
               { if (spec_func(type,ctype,arg_tp0))
                 return new denotation(shared_value 
                   (new builtin_value(suffix_element_wrapper,"#@([T],T->[T])")));
                 throw type_error(e,copy(ctype),copy(type));
               }
             }
             if (arg_tp1.kind==row_type and *arg_tp1.component_type==arg_tp0)
             { if (spec_func(type,ctype,arg_tp1))
               return new denotation(shared_value 
                 (new builtin_value(prefix_element_wrapper,"#@(T,[T]->[T])")));
               throw type_error(e,copy(ctype),copy(type));
             }
           }
         }
       }
     for (size_t i=0; i<variants.size(); ++i)
       if (variants[i].type->arg_type==ctype)
       {
         expression_ptr p(new denotation(variants[i].val));
         if (spec_func(type,ctype,variants[i].type->result_type))
           return p.release();
         type_ptr ftype=make_function_type
   	(copy(ctype),copy(variants[i].type->result_type));
         throw type_error(e,ftype,copy(type));
       }
     std::ostringstream o;
     o << "Cannot resolve " << main_hash_table->name_of(c->oper) 
       << " at argument type " << ctype;
     throw program_error(o.str());
   }
   break;
   
   case ass_stat:
   { Hash_table::id_type lhs=e.e.assign_variant->lhs;
     const expr& rhs=e.e.assign_variant->rhs;
     type_p it; expression_ptr assign; size_t i,j;
     if ((it=id_context->lookup(lhs,i,j))!=NULL)
     { expression_ptr r(convert_expr(rhs,*it));
       assign.reset(new local_assignment(lhs,i,j,r));
     }
     else if ((it=global_id_table->type_of(lhs))!=NULL)
     { expression_ptr r(convert_expr(rhs,*it));
       assign.reset(new global_assignment(lhs,r));
     }
     else throw program_error 
       (std::string("Undefined identifier in assignment: ")
        +main_hash_table->name_of(lhs));
   
     if (type.specialise(*it) or coerce(*it,type,assign))
       return assign.release();
     else throw type_error(e,copy(*it),copy(type));
   }
   break;
   
   case comp_ass_stat:
   { Hash_table::id_type aggr=e.e.comp_assign_variant->aggr;
     const expr& index=e.e.comp_assign_variant->index;
     const expr& rhs=e.e.comp_assign_variant->rhs;
   type_p aggr_t; type_expr ind_t; type_expr comp_t;
     expression_ptr assign; size_t d,o; bool is_local;
     if ((aggr_t=id_context->lookup(aggr,d,o))!=NULL)
       is_local=true;
     else if ((aggr_t=global_id_table->type_of(aggr))!=NULL)
       is_local=false;
     else throw program_error 
       (std::string("Undefined identifier in component assignment: ")
        +main_hash_table->name_of(aggr));
   
   
     expression_ptr i(convert_expr(index,ind_t));
     subscr_base::sub_type kind;
     if (subscr_base::indexable(*aggr_t,ind_t,comp_t,kind))
     { expression_ptr r(convert_expr(rhs,comp_t));
       if (is_local)
         assign.reset(new local_component_assignment(aggr,i,d,o,r,kind));
       else
         assign.reset(new global_component_assignment(aggr,i,r,kind));
     }
     else
     { std::ostringstream o;
       o << "Cannot subscript " << *aggr_t <<  " value with index of type "
         << ind_t << " in assignment";
       throw expr_error(e,o.str());
     }
   
     if (type.specialise(*aggr_t) or coerce(*aggr_t,type,assign))
       return assign.release();
     else throw type_error(e,copy(*aggr_t),copy(type));
   }
   break;
   
   case seq_expr:
   { sequence seq=e.e.sequence_variant;
     expression_ptr first(convert_expr(seq->first,void_type));
     expression_ptr last(convert_expr(seq->last,type));
     return new seq_expression(first,last);
   }
   break;
 }
 return NULL; // keep compiler happy
}

list_expression::~list_expression()
{ for (size_t i=0; i<component.size(); ++i) delete component[i];
}

void list_expression::print(std::ostream& out) const
{ out << '[';
  if (component.size()==0) out << ']';
  else
    for (size_t i=0; i<component.size(); ++i)
    out << *component[i] << (i<component.size()-1 ? ',' : ']');
}

void list_expression::evaluate(level l) const
{ if (l==no_value)
    for (size_t i=0; i<component.size(); ++i)
      component[i]->void_eval();
  else
  { row_ptr result(new row_value(component.size()));
    for (size_t i=0; i<component.size(); ++i)
      component[i]->eval(),result->val[i]=pop_value();
    push_value(result); // result will be shared from here on
  }
}

void tuple_expression::print(std::ostream& out) const
{ out << '(';
  if (component.size()==0) out << ')';
  else
    for (size_t i=0; i<component.size(); ++i)
    out << *component[i] << (i<component.size()-1 ? ',' : ')');
}

void tuple_expression::evaluate(level l) const
{ if (l==no_value)
    for (size_t i=0; i<component.size(); ++i)
      component[i]->void_eval();
  else
  { for (size_t i=0; i<component.size(); ++i)
      component[i]->eval();
    if (l==single_value)
      wrap_tuple(component.size());
  }
}

void subscr_base::print(std::ostream& out) const
{ out << *array << '[' << *index << ']';
}

void matrix_subscription::print(std::ostream& out) const
{ tuple_expression* p=dynamic_cast<tuple_expression*>(index);
  if (p==NULL) out << *array << '[' << *index << ']';
  else
    out << *array << '[' << *p->component[0] << ',' << *p->component[1] << ']';
}

bool subscr_base::indexable
  (const type_expr& aggr,
   const type_expr& index,
         type_expr& subscr,
         sub_type& kind)
{ if (aggr.kind==row_type and index==int_type)
  { kind=row_entry;
        return subscr.specialise(*aggr.component_type);
  }
  if (aggr==vec_type and index==int_type)
  { kind=vector_entry;
        return subscr.specialise(int_type);
  }
  if (aggr!=mat_type)
    return false;
  if (index==int_int_type)
  { kind=matrix_entry;
        return subscr.specialise(int_type);
  }
  kind=matrix_column;
  return index==int_type and subscr.specialise(vec_type);
}

inline std::string range_mess(int i,size_t n,const expression_base* e)
{ std::ostringstream o;
  e->print(o << "index " << i << " out of range (<" << n
             << ") in subscription ");
  return o.str();
}

void row_subscription::evaluate(level l) const
{ shared_int i=((index->eval(),get<int_value>()));
  shared_row r=((array->eval(),get<row_value>()));
  if (static_cast<unsigned int>(i->val)>=r->val.size())
    throw std::runtime_error(range_mess(i->val,r->val.size(),this));
  push_expanded(l,r->val[i->val]);
}

void vector_subscription::evaluate(level l) const
{ shared_int i=((index->eval(),get<int_value>()));
  shared_vector v=((array->eval(),get<vector_value>()));
  if (static_cast<unsigned int>(i->val)>=v->val.size())
    throw std::runtime_error(range_mess(i->val,v->val.size(),this));
  if (l!=no_value)
    push_value(new int_value(v->val[i->val]));
}

void matrix_subscription::evaluate(level l) const
{ index->multi_eval(); 
  shared_int j=get<int_value>();
  shared_int i=get<int_value>();
  shared_matrix m=((array->eval(),get<matrix_value>()));
  if (static_cast<unsigned int>(i->val)>=m->val.numRows())
    throw std::runtime_error
     ("initial "+range_mess(i->val,m->val.numRows(),this));
  if (static_cast<unsigned int>(j->val)>=m->val.numColumns())
    throw std::runtime_error
     ("final "+range_mess(j->val,m->val.numColumns(),this));
  if (l!=no_value)
    push_value(new int_value(m->val(i->val,j->val)));
}

void matrix_slice::evaluate(level l) const
{ shared_int j=((index->eval(),get<int_value>()));
  shared_matrix m=((array->eval(),get<matrix_value>()));
  if (static_cast<unsigned int>(j->val)>=m->val.numColumns())
    throw std::runtime_error(range_mess(j->val,m->val.numColumns(),this));
  if (l!=no_value)
    push_value(new vector_value(m->val.column(j->val)));
}

const char* identifier::name() const
{ return main_hash_table->name_of(code); }
void identifier::print(std::ostream& out) const
{ out<< name(); }

Id_table::~Id_table()
{ for (map_type::const_iterator p=table.begin(); p!=table.end(); ++p)
  { delete p->second.type; }
}

void Id_table::add(Hash_table::id_type id, shared_value val, type_ptr type)
{ std::pair<map_type::iterator,bool> trial=
    table.insert(std::make_pair(id,id_data()));
  id_data& slot=trial.first->second;
  if (trial.second) // a fresh global identifier
  { slot.val.reset(new shared_value(val));
    slot.type=type.release();
  }
  else
  { slot.val.reset(new shared_value(val));
    delete slot.type;
    slot.type=type.release();
  }
}

type_p Id_table::type_of(Hash_table::id_type id) const
{ map_type::const_iterator p=table.find(id);
  return p==table.end() ? NULL : p->second.type;
}
shared_value Id_table::value_of(Hash_table::id_type id) const
{ map_type::const_iterator p=table.find(id);
  return p==table.end() ? shared_value(value(NULL)) : *p->second.val;
}
shared_share Id_table::address_of(Hash_table::id_type id)
{ map_type::iterator p=table.find(id);
  if (p==table.end())
    throw std::logic_error 
    (std::string("Identifier without value:")+main_hash_table->name_of(id));

  return p->second.val;
}

void Id_table::print(std::ostream& out) const
{ for (map_type::const_iterator p=table.begin(); p!=table.end(); ++p)
    out << main_hash_table->name_of(p->first) << ": " 
        << *p->second.type << ": " << **p->second.val << std::endl;
}

std::ostream& operator<< (std::ostream& out, const Id_table& p)
{ p.print(out); return out; }

global_identifier::global_identifier(Hash_table::id_type id)
: identifier(id), address(global_id_table->address_of(id))
{}

void global_identifier::evaluate(level l) const
{ if (address->get()==NULL)
  { std::ostringstream o;
    o << "Taking value of uninitialized variable " << name();
    throw std::runtime_error(o.str());
  }
  push_expanded(l,*address);
}

void local_identifier::evaluate(level l) const
{ push_expanded(l,execution_context->elem(depth,offset)); }

overload_table::~overload_table()
{ for (map_type::const_iterator p=table.begin(); p!=table.end(); ++p)
    for (size_t i=0; i<p->second.size(); ++i)
    { delete p->second[i].type; }
}

const overload_table::variant_list& overload_table::variants
  (Hash_table::id_type id) const
{ static const variant_list empty;
  map_type::const_iterator p=table.find(id);
  return p==table.end() ? empty : p->second;
}

void overload_table::add
  (Hash_table::id_type id, shared_value val, type_ptr tp)
{ if (tp->kind!=function_type)
    throw std::logic_error("Wrapper function has non-function type");
  func_type_ptr type(tp->func);
  tp->kind=undetermined_type; // release
  if (type->arg_type==void_type)
    throw program_error("Cannot define function overload without arguments");
  std::pair<map_type::iterator,bool> trial=
    table.insert(std::make_pair(id,variant_list(1,overload_data())));
  variant_list& slot=trial.first->second;
  if (trial.second) // a fresh overloaded identifier
  { slot[0].val=val;
    slot[0].type=type.release();
  }
  else
  
  { size_t lwb=0; size_t upb=slot.size();
    for (size_t i=0; i<slot.size(); ++i)
    { unsigned int cmp= is_close(type->arg_type,slot[i].type->arg_type);
      switch (cmp)
      {
        case 0x6: lwb=i+1; break; // existent type |i| converts to |type|
        case 0x5:  if (upb>i) upb=i;  break; // |type| converts to type |i|
        case 0x7:
          if (slot[i].type->arg_type==type->arg_type)
          { slot[i].val=val;
              delete slot[i].type;
              slot[i].type=type.release();
              return;
            }
        // |else| {\bf fall through}
        case 0x4: // conflicting cases
          { std::ostringstream o;
            o << "Cannot overload `" << main_hash_table->name_of(id) << "', " 
              << "existing type\n" << slot[i].type->arg_type
              << " is close to " << type->arg_type << ",\n"
              << (cmp==0x4 ? "but neither" :"and either") 
              << " converts to the other";
            throw program_error(o.str());
          }
        default: {} // nothing for unrelated argument types
      }
    }
    if(lwb>upb)
      throw std::logic_error("Conflicting order related types");
    
    { slot.insert(slot.begin()+upb,overload_data());
    slot[upb].val=val; slot[upb].type=type.release();
    }
  }
}

bool overload_table::remove(Hash_table::id_type id, const type_expr& arg_t)
{ map_type::iterator p=table.find(id);
  if (p==table.end()) return false;
  variant_list& variants=p->second;
  for (size_t i=0; i<variants.size(); ++i)
    if (variants[i].type->arg_type==arg_t)
    { delete variants[i].type;
      variants.erase(variants.begin()+i);
      return true;
    }
  return false;
}

void overload_table::print(std::ostream& out) const
{ type_expr type; type.kind=function_type;
  for (map_type::const_iterator p=table.begin(); p!=table.end(); ++p)
    for (size_t i=0; i<p->second.size(); ++i)
    { type.func = p->second[i].type;
      out << main_hash_table->name_of(p->first) << ": " 
        << type << ": " << *p->second[i].val << std::endl;
    }
  type.kind=undetermined_type; // avoid destruction of |type.func|
}

std::ostream& operator<< (std::ostream& out, const overload_table& p)
{ p.print(out); return out; }

expression resolve_overload
  (const expr& e,
   type_expr& type,
   const overload_table::variant_list& variants)
{ const expr& args = e.e.call_variant->arg;
  type_expr a_priori_type;
  expression_ptr arg(convert_expr(args,a_priori_type));
  Hash_table::id_type id =  e.e.call_variant->fun.e.identifier_variant;
  
  { if (id==size_of_name())
    { if (a_priori_type.kind==row_type)
      { expression_ptr call(new overloaded_builtin_call(sizeof_wrapper,"#",arg));
        if (type.specialise(int_type) or coerce(int_type,type,call))
          return call.release();
        throw type_error(e,copy(type),copy(int_type));
      }
      else if (a_priori_type.kind!=undetermined_type and
               a_priori_type.specialise(pair_type))
      
      { type_expr& arg_tp0 = a_priori_type.tuple->t;
        type_expr& arg_tp1 = a_priori_type.tuple->next->t;
        if (arg_tp0.kind==row_type)
        { if (arg_tp0.component_type->specialise(arg_tp1) or 
              can_coerce_arg(arg.get(),1,arg_tp1,*arg_tp0.component_type)) // suffix
          { expression_ptr call(new overloaded_builtin_call
              (suffix_element_wrapper,"#",arg));
            if (type.specialise(arg_tp0) or coerce(arg_tp0,type,call))
              return call.release();
            throw type_error(e,copy(type),copy(arg_tp0));
          }
          if (arg_tp0==arg_tp1) // join
          { expression_ptr call(new overloaded_builtin_call
              (join_rows_wrapper,"#",arg));
            if (type.specialise(arg_tp0) or coerce(arg_tp0,type,call))
              return call.release();
            throw type_error(e,copy(type),copy(arg_tp0));
          }
        }
        if (arg_tp1.kind==row_type and 
               (arg_tp1.component_type->specialise(arg_tp0) or 
                can_coerce_arg(arg.get(),0,arg_tp0,*arg_tp1.component_type)))
                // prefix
        { expression_ptr call(new overloaded_builtin_call
            (prefix_element_wrapper,"#",arg));
          if (type.specialise(arg_tp1) or coerce(arg_tp1,type,call))
            return call.release();
          throw type_error(e,copy(type),copy(arg_tp1));
        }
      }
    }
    else if (id==print_name() or id==prints_name()) // these always match
    { expression c = id==print_name() 
        ? new generic_builtin_call(print_wrapper,"print",arg) 
        : new generic_builtin_call(prints_wrapper,"prints",arg);
      expression_ptr call(c); // get ownership
      if (type.specialise(void_type))
        return call.release();
      throw type_error(e,copy(type),copy(void_type));
    }
  
  }
  for (size_t i=0; i<variants.size(); ++i)
  { const overload_data& v=variants[i];
    if ((is_close(a_priori_type,v.type->arg_type)&0x1)!=0)
      // could first convert to second?
    { if (a_priori_type!=v.type->arg_type)
        arg=expression_ptr(convert_expr(args,v.type->arg_type));
          // redo conversion
      
      { expression_ptr call;
        builtin_value* f = dynamic_cast<builtin_value*>(v.val.get());
        if (f!=NULL)
          call = expression_ptr
            (new overloaded_builtin_call(f->val,f->print_name.c_str(),arg));
        else 
             { shared_closure fun =
                std::tr1::dynamic_pointer_cast<closure_value>(v.val);
               if (fun==NULL)
                 throw std::logic_error("Overloaded value is not a function");
               std::ostringstream name;
               name << main_hash_table->name_of(id) << '@' << v.type->arg_type;
               call = expression_ptr (new overloaded_closure_call(fun,name.str(),arg));
             }
        if (type.specialise(v.type->result_type) or
              coerce(v.type->result_type,type,call))
          return call.release();
        throw type_error(e,copy(v.type->result_type),copy(type));
              // result type mismatch
       }
    }
  }

  
  if (variants.size()==1)
    throw type_error(args,copy(a_priori_type),copy(variants[0].type->arg_type));
  else
  { std::ostringstream o;
    o << "Failed to match `"
      << main_hash_table->name_of(id) 
      << "' with argument type "
      << a_priori_type;
    throw expr_error(e,o.str());
  }
}

void call_expression::print(std::ostream& out) const
{ if (dynamic_cast<identifier*>(function)!=NULL) out << *function;
  else out << '(' << *function << ')';
  if (dynamic_cast<tuple_expression*>(argument)!=NULL) out << *argument;
  else out << '(' << *argument << ')';
}

void overloaded_builtin_call::print(std::ostream& out) const
{ out << print_name;
  if (dynamic_cast<tuple_expression*>(argument)!=NULL) out << *argument;
  else out << '(' << *argument << ')';
}

void call_expression::evaluate(level l) const
{ function->eval();  shared_value fun=pop_value();
builtin_value* f=dynamic_cast<builtin_value*>(fun.get());
  argument->evaluate(f==NULL ? single_value : multi_value);
  try
  { if (f==NULL)
      
      { closure_value* f=force<closure_value>(fun.get());
      std::vector<shared_value> new_frame;
        new_frame.reserve(count_identifiers(*f->param));
        thread_components(*f->param,pop_value(),new_frame);
      
        context_ptr saved_context(execution_context);
        execution_context.reset(new context(f->cont,new_frame));
        f->body->evaluate(l); // pass evaluation level |l| to function body
        execution_context = saved_context;
      }
    else // built-in functions
      (*f->val)(l); // call the wrapper function, handling |l| appropriately
  }
  
  catch (const std::exception& e)
  { identifier* p=dynamic_cast<identifier*>(function);
    if (p!=NULL) // named function
    {
      const std::logic_error* l_err= dynamic_cast<const std::logic_error*>(&e);
      if (l_err!=NULL)
        throw std::logic_error
          (std::string(e.what())+"\n(in call of "+p->name()+')');
      throw std::runtime_error
        (std::string(e.what())+"\n(in call of "+p->name()+')');
    }
    throw; // for anonymous function calls, just rethrow the error unchanged
  }
}

void overloaded_builtin_call::evaluate(level l) const
{ argument->multi_eval();
  try
  { (*f)(l); }
  catch (const std::exception& e)
  { const std::logic_error* l_err= dynamic_cast<const std::logic_error*>(&e);
    if (l_err!=NULL)
      throw std::logic_error
        (std::string(e.what())+"\n(in call of "+print_name+')');
    throw std::runtime_error
      (std::string(e.what())+"\n(in call of "+print_name+')');
  }
}

void generic_builtin_call::evaluate(level l) const
{ argument->eval();
  try
  { (*f)(l); }
  catch (const std::exception& e)
  { const std::logic_error* l_err= dynamic_cast<const std::logic_error*>(&e);
    if (l_err!=NULL)
      throw std::logic_error
        (std::string(e.what())+"\n(in call of "+print_name+')');
    throw std::runtime_error
      (std::string(e.what())+"\n(in call of "+print_name+')');
  }
}

inline
lambda_expression::lambda_expression(const id_pat& p, expression_ptr b)
: param(new id_pat(copy_id_pat(p)),id_pat_deleter()) , body(b.release())
{}

void lambda_expression::print(std::ostream& out) const
{ if (param.get()==NULL)
    out << "()";
  else if ((param->kind&0x1)!=0)
    out << '(' << *param << ')';
  else out << *param;
  out << ": " << *body;
}

void closure_value::print(std::ostream& out) const
{ if (param.get()==NULL)
    out << "()";
  else if ((param->kind&0x1)!=0)
    out << '(' << *param << ')';
  else out << *param;
  out << ": " << *body;
}

void lambda_expression::evaluate(level l) const
{ if (l!=no_value)
  {closure_ptr result(new closure_value(execution_context,param,body));
     push_value(result);
  }
}

void overloaded_closure_call::print(std::ostream& out) const
{ out << print_name;
  if (dynamic_cast<tuple_expression*>(argument)!=NULL) out << *argument;
  else out << '(' << *argument << ')';
}

void overloaded_closure_call::evaluate(level l) const
{ argument->eval();
  try
  { std::vector<shared_value> new_frame;
    new_frame.reserve(count_identifiers(*fun->param));
    thread_components(*fun->param,pop_value(),new_frame);

    context_ptr saved_context(execution_context);
    execution_context.reset(new context(fun->cont,new_frame));
    fun->body->evaluate(l); // pass evaluation level |l| to function body
    execution_context = saved_context;
  }
  catch (const std::exception& e)
  { const std::logic_error* l_err= dynamic_cast<const std::logic_error*>(&e);
    if (l_err!=NULL)
      throw std::logic_error
        (std::string(e.what())+"\n(in call of "+print_name+')');
    throw std::runtime_error
      (std::string(e.what())+"\n(in call of "+print_name+')');
  }
}

void conditional_expression::print(std::ostream& out) const
{ out << " if "; const conditional_expression* cur=this;
  do
  { out << *cur->condition << " then " << *cur->then_branch;
    conditional_expression* p =
      dynamic_cast<conditional_expression*>(cur->else_branch);
    if (p==NULL)
      break;
    out << " elif "; cur=p;
  }
  while(true); // \Cpp\ does not allow using |p| in final condition
  out << " else " << *cur->else_branch << " fi ";
}

void conditional_expression::evaluate(level l) const
{ condition->eval();
  if (get<bool_value>()->val)
   then_branch->evaluate(l);
  else else_branch->evaluate(l);
}

void while_expression::print(std::ostream& out) const
{ out << " while " <<  *condition << " do " << *body;
  if (next_part!=NULL)
    out << " next " << *next_part;
  out << " od ";
}

void while_expression::evaluate(level l) const
{ if (l==no_value)
    while (condition->eval(),get<bool_value>()->val)
    { body->void_eval();
      if (next_part!=NULL)
        next_part->void_eval();
    }
  else
  { row_ptr result (new row_value(0));
    while (condition->eval(),get<bool_value>()->val)
    { body->eval(); result->val.push_back(pop_value());
      if (next_part!=NULL)
        next_part->void_eval();
    }
    push_value(result);
  }
}

inline
for_expression::for_expression
 (id_pat p, expression_ptr i, expression_ptr b,subscr_base::sub_type k)
   : pattern(copy_id_pat(p)), in_part(i.release()), body(b.release()), kind(k)
  {}

void for_expression::print(std::ostream& out) const
{ out << " for " << pattern.sublist->next->body;
    if (pattern.sublist->body.kind==0x1)
      out << '@' << pattern.sublist->body;
    out << " in " << *in_part << " do " << *body << " od ";
}

void for_expression::evaluate(level l) const
{ size_t n_id = count_identifiers(pattern);
  context_ptr saved_context=execution_context;
  in_part->eval();
  shared_tuple loop_var(new tuple_value(2));
       // this is safe to re-use between iterations
  std::vector<shared_value> loop_frame; loop_frame.reserve(n_id);

  row_ptr result(NULL);
  if (kind==subscr_base::row_entry)
  { shared_row in_val = get<row_value>();
    size_t n=in_val->val.size();
    if (l!=no_value)
      result = row_ptr(new row_value(n));
    for (size_t i=0; unsigned(i)<n; ++i,loop_frame.clear())
    { loop_var->val[1]=in_val->val[i];
      
      loop_var->val[0].reset(new int_value(i)); // must be newly created each time
      thread_components(pattern,loop_var,loop_frame);
      execution_context.reset(new context(saved_context,loop_frame));
        // this one too
      if (l==no_value)
        body->void_eval();
      else
      { body->eval(); result->val[i]=pop_value(); }
    }
  }
  else if (kind==subscr_base::vector_entry)
  { shared_vector in_val = get<vector_value>();
    size_t n=in_val->val.size();
    if (l!=no_value)
      result = row_ptr(new row_value(n));
    for (size_t i=0; unsigned(i)<n; ++i,loop_frame.clear())
    { loop_var->val[1].reset(new int_value(in_val->val[i]));
      
      loop_var->val[0].reset(new int_value(i)); // must be newly created each time
      thread_components(pattern,loop_var,loop_frame);
      execution_context.reset(new context(saved_context,loop_frame));
        // this one too
      if (l==no_value)
        body->void_eval();
      else
      { body->eval(); result->val[i]=pop_value(); }
    }
  }
  else if (kind==subscr_base::matrix_column)
  { shared_matrix in_val = get<matrix_value>();
    size_t n=in_val->val.numColumns();
    if (l!=no_value)
      result = row_ptr(new row_value(n));
    for (size_t i=0; unsigned(i)<n; ++i,loop_frame.clear())
    { loop_var->val[1].reset(new vector_value(in_val->val.column(i)));
      
      loop_var->val[0].reset(new int_value(i)); // must be newly created each time
      thread_components(pattern,loop_var,loop_frame);
      execution_context.reset(new context(saved_context,loop_frame));
        // this one too
      if (l==no_value)
        body->void_eval();
      else
      { body->eval(); result->val[i]=pop_value(); }
    }
  }
  execution_context = saved_context;
  if (l!=no_value)
    push_value(result);
}

void inc_for_expression::print(std::ostream& out) const
{ out << " for " << main_hash_table->name_of(id)  << " = " << *count 
      << " from " << *bound << " do " << *body << " od ";
}
void dec_for_expression::print(std::ostream& out) const
{ out << " for " << main_hash_table->name_of(id)  << " = " << *count 
      << " downto " << *bound << " do " << *body << " od ";
}

void inc_for_expression::evaluate(level l) const
{ int b=(bound->eval(),get<int_value>()->val);
  int c=(count->eval(),get<int_value>()->val);
  if (c<0)
    c=0; // no negative size result

  context_ptr saved_context=execution_context;
  std::vector<shared_value>loop_frame(1);
  shared_value& loop_var=loop_frame[0];
  if (l==no_value)
  { c+=b;
    for (int i=b; i<c; ++i)
    { loop_var.reset(new int_value(i));
      execution_context.reset(new context(saved_context,loop_frame));
      body->void_eval();
    }
  }
  else
  { row_ptr result (new row_value(0)); result->val.reserve(c);
    c+=b;
    for (int i=b; i<c; ++i)
    { loop_var.reset(new int_value(i));
      execution_context.reset(new context(saved_context,loop_frame));
      body->eval(); result->val.push_back(pop_value());
    }
    push_value(result);
  }
  execution_context=saved_context;
}

void dec_for_expression::evaluate(level l) const
{ int b=(bound->eval(),get<int_value>()->val);
  int i=(count->eval(),get<int_value>()->val);
  if (i<0)
    i=0; // no negative size result

  context_ptr saved_context=execution_context;
  std::vector<shared_value>loop_frame(1);
  shared_value& loop_var=loop_frame[0];
  if (l==no_value)
  { i+=b;
    while (i-->b)
    { loop_var.reset(new int_value(i));
      execution_context.reset(new context(saved_context,loop_frame));
      body->void_eval();
    }
  }
  else
  { row_ptr result (new row_value(0)); result->val.reserve(i);
    i+=b;
    while (i-->b)
    { loop_var.reset(new int_value(i));
      execution_context.reset(new context(saved_context,loop_frame));
      body->eval(); result->val.push_back(pop_value());
    }
    push_value(result);
  }
  execution_context=saved_context;
}

void assignment_expr::print(std::ostream& out) const
{ out << main_hash_table->name_of(lhs) << ":=" << *rhs; }

global_assignment::global_assignment(Hash_table::id_type l,expression_ptr r)
: assignment_expr(l,r), address(global_id_table->address_of(l)) {}

void global_assignment::evaluate(level l) const
{ rhs->eval();
  *address = pop_value();
  push_expanded(l,*address);
}

local_assignment::local_assignment
 (Hash_table::id_type l, size_t i,size_t j, expression_ptr r)
: assignment_expr(l,r), depth(i), offset(j) {}

void local_assignment::evaluate(level l) const
{ rhs->eval();
  shared_value& dest =  execution_context->elem(depth,offset);
  dest= pop_value();
  push_expanded(l,dest);
}

void component_assignment::print(std::ostream& out) const
{ out << main_hash_table->name_of(lhs) << '[' << *index << "]:=" << *rhs; }

global_component_assignment::global_component_assignment 
  (Hash_table::id_type a,expression_ptr i,expression_ptr r,
   subscr_base::sub_type k)
: component_assignment(a,i,r)
, kind(k),address(global_id_table->address_of(a)) {}

void global_component_assignment::evaluate(level l) const
{ if (address->get()==NULL)
  { std::ostringstream o;
    o << "Assigning to component of uninitialized variable "
      << main_hash_table->name_of(lhs);
    throw std::runtime_error(o.str());
  }
  assign(l,*address,kind);
}

void component_assignment::assign
  (level l,shared_value& aggregate, subscr_base::sub_type kind) const
{ rhs->eval();
  uniquify(aggregate);
  value loc=aggregate.get(); // simple reference from shared pointer
  switch (kind)
  { case subscr_base::row_entry:
  
  { unsigned int i=(index->eval(),get<int_value>()->val);
    row_value& a=*force<row_value>(loc);
    if (i>=a.val.size())
      throw std::runtime_error(range_mess(i,a.val.size(),this));
    a.val[i]= pop_value();
    push_expanded(l,a.val[i]);
  }
  break;
    case subscr_base::vector_entry:
  
  { unsigned int i=(index->eval(),get<int_value>()->val);
    vector_value& v=*force<vector_value>(loc);
    if (i>=v.val.size())
      throw std::runtime_error(range_mess(i,v.val.size(),this));
    v.val[i]= force<int_value>(execution_stack.back().get())->val;
    if (l==no_value)
      execution_stack.pop_back();
  }
  break;
    case subscr_base::matrix_entry:
  
  { index->multi_eval();
    unsigned int j=get<int_value>()->val;
    unsigned int i=get<int_value>()->val;
  
    matrix_value& m=*force<matrix_value>(loc);
    if (i>=m.val.numRows())
      throw std::runtime_error(range_mess(i,m.val.numRows(),this));
    if (j>=m.val.numColumns())
      throw std::runtime_error(range_mess(j,m.val.numColumns(),this));
    m.val(i,j)= force<int_value>(execution_stack.back().get())->val;
    if (l==no_value)
      execution_stack.pop_back();
  }
  break;
    case subscr_base::matrix_column:
  
  { unsigned int j=(index->eval(),get<int_value>()->val);
    matrix_value& m=*force<matrix_value>(loc);
    const vector_value& v=*force<vector_value>(execution_stack.back().get());
    if (j>=m.val.numColumns())
      throw std::runtime_error(range_mess(j,m.val.numColumns(),this));
    if (v.val.size()!=m.val.numRows())
      throw std::runtime_error
        (std::string("Cannot replace column of size ")+str(m.val.numRows())+
         " by one of size "+str(v.val.size()));
    m.val.set_column(j,v.val);
    if (l==no_value)
      execution_stack.pop_back();
  }
  break;
  }
}

local_component_assignment::local_component_assignment
 (Hash_table::id_type l, expression_ptr i,size_t d, size_t o, expression_ptr r,
  subscr_base::sub_type k)
: component_assignment(l,i,r), kind(k), depth(d), offset(o) {}

void local_component_assignment::evaluate(level l) const
{ assign(l,execution_context->elem(depth,offset),kind); }

void seq_expression::print(std::ostream& out) const
{ out << *first << ';' << *last; }

void seq_expression::evaluate(level l) const
{ first->void_eval(); last->evaluate(l); }

type_ptr analyse_types(const expr& e,expression_ptr& p)
  throw(std::bad_alloc,std::runtime_error)
{ try
  { type_ptr type=copy(unknown_type);
    p.reset(convert_expr(e,*type));
    return type;
  }
  catch (type_error& err)
  { std::cerr << err.what() << ":\n  Subexpression " << err.offender << 
    " has wrong type: found " << *err.actual << 
    " while " << *err.required << " was needed.\n";

  }
  catch (expr_error& err)
  { std::cerr << "Error in expression " << err.offender << "\n  "
              << err.what() << std::endl;
  }
  catch (program_error& err)
  { std::cerr << "Error during analysis of expression " << e << "\n  "
                << err.what() << std::endl;
  }
  throw std::runtime_error("Type check failed");

}

void vector_value::print(std::ostream& out) const
{ size_t l=val.size(),w=0; std::vector<std::string> tmp(l);
  for (size_t i=0; i<l; ++i)
  { std::ostringstream s; s<<val[i]; tmp[i]=s.str();
    if (tmp[i].length()>w) w=tmp[i].length();
  }
  if (l==0) out << "[ ]";
  else
  { w+=1; out << std::right << '[';
    for (size_t i=0; i<l; ++i)
      out << std::setw(w) << tmp[i] << (i<l-1 ? "," : " ]");
  }
}

void rational_vector_value::print(std::ostream& out) const
{ size_t l=val.size(),w=0; std::vector<std::string> tmp(l);
  for (size_t i=0; i<l; ++i)
  { std::ostringstream s; s<<val.numerator()[i]; tmp[i]=s.str();
    if (tmp[i].length()>w) w=tmp[i].length();
  }
  if (l==0) out << "[ ]";
  else
  { w+=1; out << std::right << '[';
    for (size_t i=0; i<l; ++i)
      out << std::setw(w) << tmp[i] << (i<l-1 ? "," : " ]");
  }
  out << '/' << val.denominator();
}

void matrix_value::print(std::ostream& out) const
{ size_t k=val.numRows(),l=val.numColumns();
  if (k==0 or l==0)
  {  out << "The unique " << k << 'x' << l << " matrix"; return; }
  std::vector<size_t> w(l,0);
  for (size_t i=0; i<k; ++i)
    for (size_t j=0; j<l; ++j)
    { std::ostringstream s; s<<val(i,j); size_t len=s.str().length();
      if (len>w[j]) w[j]=len;
    }
  out << std::endl << std::right;
  for (size_t i=0; i<k; ++i)
  { out << '|';
    for (size_t j=0; j<l; ++j)
      out << std::setw(w[j]+1) << val(i,j) << (j<l-1 ? ',' : ' ');
    out << '|' << std::endl;
  }
}

void install_function
 (wrapper_function f,const char*name, const char* type_string)
{ type_ptr type = make_type(type_string);
  std::ostringstream print_name; print_name<<name;
  if (type->kind!=function_type)
    throw std::logic_error
     ("Built-in with non-function type: "+print_name.str());
  if (type->func->arg_type==void_type)
  { shared_value val(new builtin_value(f,print_name.str()));
    global_id_table->add (main_hash_table->match_literal(name),val,type);
  }
  else
  { print_name << '@' << type->func->arg_type;
    shared_value val(new builtin_value(f,print_name.str()));
    global_overload_table->add(main_hash_table->match_literal(name),val,type);
  }
}

void id_mat_wrapper(expression_base::level l)
{ int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
  { matrix_ptr m (new matrix_value(matrix::Matrix<int>()));
    matrix::identityMatrix(m->val,std::abs(i)); push_value(m);
  }
}
 void transpose_mat_wrapper(expression_base::level l)
{ matrix::Matrix<int> m=get<matrix_value>()->val;
  if (l!=expression_base::no_value)
    push_value(new matrix_value(m.transposed()));
}
 void transpose_vec_wrapper(expression_base::level l)
{ matrix::Vector<int> v=get<vector_value>()->val;
  if (l!=expression_base::no_value)
  { matrix_ptr m (new matrix_value(matrix::Matrix<int>(1,v.size())));
    for (size_t j=0; j<v.size(); ++j)
      m->val(0,j)=v[j];
    push_value(m);
  }
}

void vv_prod_wrapper(expression_base::level l)
{ matrix::Vector<int> w=get<vector_value>()->val;
  matrix::Vector<int> v=get<vector_value>()->val;
  if (v.size()!=w.size())
    throw std::runtime_error(std::string("Size mismatch ")
     + str(v.size()) + ":" + str(w.size()) + " in scalar product");
  if (l!=expression_base::no_value)
    push_value(new int_value(v.dot(w)));
}

void mv_prod_wrapper(expression_base::level l)
{ matrix::Vector<int> v=get<vector_value>()->val;
  matrix::Matrix<int> m=get<matrix_value>()->val;
  if (m.numColumns()!=v.size())
    throw std::runtime_error(std::string("Size mismatch ")
     + str(m.numColumns()) + ":" + str(v.size()) + " in mat*vec");
  if (l!=expression_base::no_value)
    push_value(new vector_value(m.apply(v)));
}

void mm_prod_wrapper(expression_base::level l)
{ matrix::Matrix<int> rf=get<matrix_value>()->val; // right factor
  matrix::Matrix<int> lf=get<matrix_value>()->val; // left factor
  if (lf.numColumns()!=rf.numRows())
  { std::ostringstream s;
    s<< "Size mismatch " << lf.numColumns() << ":" << rf.numRows()
     << " in mat*mat";
    throw std::runtime_error(s.str());
  }
  if (l!=expression_base::no_value)
    push_value(new matrix_value(lf*rf));
}

void vm_prod_wrapper(expression_base::level l)
{ matrix::Matrix<int> m=get<matrix_value>()->val; // right factor
  matrix::Vector<int> v=get<vector_value>()->val; // left factor
  if (v.size()!=m.numRows())
  { std::ostringstream s;
    s<< "Size mismatch " << v.size() << ":" << m.numRows()
     << " in vec*mat";
    throw std::runtime_error(s.str());
  }
  if (l!=expression_base::no_value)
    push_value(new vector_value(m.right_apply(v)));
}

extern "C"
void global_set_identifier(id_pat pat, expr rhs, int overload)
{ using namespace atlas::interpreter;
  size_t n_id=count_identifiers(pat);
  int phase=0;
  static const char* phase_name[3] = {"type_check","evaluation","definition"};
  try
  { expression_ptr e;
    type_ptr t=analyse_types(rhs,e);
    if (not pattern_type(pat)->specialise(*t))
      
      { std::ostringstream o;
        o << "Type " << *t 
          << " of right hand side does not match required pattern "
          << *pattern_type(pat);
        throw std::runtime_error(o.str());
      }

    phase=1;
    bindings b(n_id);
    thread_bindings(pat,*t,b); // match identifiers and their future types

    std::vector<shared_value> v;
    v.reserve(n_id);
  e->eval();
    thread_components(pat,pop_value(),v);

    phase=2;
    if (overload==0)
      
      { if (n_id>0)
          std::cout << "Identifier";
        for (size_t i=0; i<n_id; ++i)
        { std::cout << (i==0 ? n_id==1 ? " " : "s " : ", ") 
                    << main_hash_table->name_of(b[i].first) << ": "
                    << *b[i].second;
          global_id_table->add(b[i].first,v[i],copy(*b[i].second));
        }
      }
    else
      
      { assert(n_id=1);
        size_t old_n=global_overload_table->variants(b[0].first).size();
        global_overload_table->add(b[0].first,v[0],copy(*b[0].second));
        size_t n=global_overload_table->variants(b[0].first).size();
        if (n==old_n)
          std::cout << "Redefined ";
        else if (n==1)
          std::cout << "Defined ";
        else
          std::cout << "Added definition [" << n << "] of ";
        std::cout << main_hash_table->name_of(b[0].first) << ": " << *b[0].second;
      }
    std::cout << std::endl;
  }
  
  catch (std::runtime_error& err)
  { std::cerr << err.what() << '\n';
    if (n_id>0)
    { std::vector<Hash_table::id_type> names; names.reserve(n_id);
      list_identifiers(pat,names);
      std::cerr << "  Identifier" << (n_id==1 ? "" : "s");
      for (size_t i=0; i<n_id; ++i)
        std::cerr << (i==0 ? " " : ", ") << main_hash_table->name_of(names[i]);
      std::cerr << " not " << (overload==0 ? "created." : "overloaded.")
                << std::endl;
    }
    reset_evaluator(); main_input_buffer->close_includes();
  }
  catch (std::logic_error& err)
  { std::cerr << "Unexpected error: " << err.what() << ", " 
              << phase_name[phase]
              << " aborted.\n";
  reset_evaluator(); main_input_buffer->close_includes();
  }
  catch (std::exception& err)
  { std::cerr << err.what() << ", "
              << phase_name[phase]
              << " aborted.\n";
  reset_evaluator(); main_input_buffer->close_includes();
  }
}

extern "C"
void global_declare_identifier(Hash_table::id_type id, ptr t)
{ value undef=NULL;
  const type_expr& type=*static_cast<type_p>(t);
  global_id_table->add(id,shared_value(undef),copy(type));
  std::cout << "Identifier " << main_hash_table->name_of(id)
            << " : " << type << std::endl;
}

extern "C"
void type_of_expr(expr e)
{ try
  { expression_ptr p;
    *output_stream << "type: " << *analyse_types(e,p) << std::endl;
  }
  catch (std::exception& err) { std::cerr<<err.what()<<std::endl; }
}

extern "C"
void show_overloads(id_type id)
{ const overload_table::variant_list& variants =
   global_overload_table->variants(id);
  *output_stream
   << (variants.empty() ? "No overloads for " : "Overloaded instances of ") 
   << main_hash_table->name_of(id) << std::endl;
 for (size_t i=0; i<variants.size(); ++i)
   *output_stream << "  "
    << variants[i].type->arg_type << "->" << variants[i].type->result_type 
    << std::endl;
}

extern "C"
void show_ids()
{ *output_stream << "Overloaded operators and functions:\n"
                 << *global_overload_table 
                 << "Global values:\n" << *global_id_table;
}

} }

