
#ifndef EVALUATOR_H
#define EVALUATOR_H

#include "types.h"


#include <string>
#include <sstream>

#include "arithmetic.h"

#include <map>
#include "lexer.h" 

#include "latticetypes.h"

namespace atlas { namespace interpreter {

struct int_value : public value_base
{ int val;

  explicit int_value(int v) : val(v)  {}
  ~int_value() {}
  void print(std::ostream& out) const { out << val; }
  int_value* clone() const { return new int_value(*this); }
  static const char* name() { return "integer"; }
private:
  int_value(const int_value& v) : val(v.val) {}
};

typedef std::auto_ptr<int_value> int_ptr;
typedef std::tr1::shared_ptr<int_value> shared_int;

struct rat_value : public value_base
{ arithmetic::Rational val;

  explicit rat_value(arithmetic::Rational v) : val(v)  {}
  ~rat_value() {}
  void print(std::ostream& out) const { out << val; }
  rat_value* clone() const { return new rat_value(*this); }
  static const char* name() { return "integer"; }
private:
  rat_value(const rat_value& v) : val(v.val) {}
};

typedef std::auto_ptr<rat_value> rat_ptr;
typedef std::tr1::shared_ptr<rat_value> shared_rat;

struct string_value : public value_base
{ std::string val;

  explicit string_value(const std::string& s) : val(s)  {}
  ~string_value() {}
  void print(std::ostream& out) const { out << '"' << val << '"'; }
  string_value* clone() const { return new string_value(*this); }
  static const char* name() { return "string"; }
private:
  string_value(const string_value& v) : val(v.val) {}
};

typedef std::auto_ptr<string_value> string_ptr;
typedef std::tr1::shared_ptr<string_value> shared_string;


struct bool_value : public value_base
{ bool val;

  explicit bool_value(bool v) : val(v)  {}
  ~bool_value() {}
  void print(std::ostream& out) const { out << std::boolalpha << val; }
  bool_value* clone() const { return new bool_value(*this); }
  static const char* name() { return "Boolean"; }
private:
  bool_value(const bool_value& v) : val(v.val) {}
};

typedef std::auto_ptr<bool_value> bool_ptr;
typedef std::tr1::shared_ptr<bool_value> shared_bool;

struct denotation : public expression_base
{ shared_value denoted_value;

  explicit denotation(shared_value v) : denoted_value(v) {}
  virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const
  { denoted_value->print(out); }
};

class bindings
: public std::vector<std::pair<Hash_table::id_type,type_p> >
{ typedef std::vector<std::pair<Hash_table::id_type,type_p> >
    base;
  bindings* next; // non-owned pointer
bindings(const bindings&); // copy constructor
  void operator= (const bindings&); // assignment operator
public:
  bindings(size_t n=0) :  base() , next(NULL) { base::reserve(n); }
    // predict size |n|, informative
  ~bindings () 
  { for (base::iterator it=begin(); it!=end(); ++it)
      delete it->second;
  }
  void add(Hash_table::id_type id,type_ptr t);
  type_p lookup
    (Hash_table::id_type id, size_t& depth, size_t& offset) const;
  void push (bindings*& sp) { next=sp; sp=this; }
  void pop (bindings*& sp) const { sp=next; }
};

struct list_expression : public expression_base
{ std::vector<expression> component;

  explicit list_expression(size_t n) : component(n,NULL) {}
   // always start out with null pointers
  virtual ~list_expression();
  virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const;
};

struct tuple_expression : public list_expression
{ explicit tuple_expression(size_t n) : list_expression(n){}
  virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const;
};

struct subscr_base : public expression_base
{ enum sub_type { row_entry, vector_entry, matrix_entry, matrix_column };
  expression array, index;

  subscr_base(expression_ptr a, expression_ptr i)
  : array(a.release()),index(i.release()) {}
  ~subscr_base()  { delete array; delete index; }
  void print(std::ostream& out) const;
  static bool indexable
  (const type_expr& aggr,
   const type_expr& index,
         type_expr& subscr,
         sub_type& kind);
};


struct row_subscription : public subscr_base
{ row_subscription(expression_ptr a, expression_ptr i)
  : subscr_base(a,i) {}
  virtual void evaluate(level l) const;
};


struct vector_subscription : public subscr_base
{ vector_subscription(expression_ptr a, expression_ptr i)
  : subscr_base(a,i) {}
  virtual void evaluate(level l) const;
};

struct matrix_subscription : public subscr_base
{ matrix_subscription(expression_ptr a, expression_ptr ij)
  : subscr_base(a,ij) {}
  virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const;
};

struct matrix_slice : public subscr_base
{ matrix_slice(expression_ptr a, expression_ptr j)
  : subscr_base(a,j) {}
  virtual void evaluate(level l) const;
};

struct identifier : public expression_base
{ Hash_table::id_type code;

  explicit identifier(Hash_table::id_type id) : code(id) {}
  virtual ~identifier()  {}
  const char* name() const;
  virtual void print(std::ostream& out) const;
};

typedef std::tr1::shared_ptr<shared_value> shared_share;
struct id_data
{ shared_share val;  type_p type;
  id_data() : val(),type(NULL) {}
};

class Id_table
{ typedef std::map<Hash_table::id_type,id_data> map_type;
  map_type table;
  Id_table(const Id_table&); // copying forbidden
  Id_table& operator=(const Id_table&); // assignment forbidden
public:
  Id_table() : table() {} // the default and only accessible constructor
  ~Id_table(); // destructor of all values referenced in the table

  void add(Hash_table::id_type id, shared_value v, type_ptr t); // insertion
  type_p type_of(Hash_table::id_type id) const; // lookup
  shared_value value_of(Hash_table::id_type id) const; // lookup
  shared_share address_of(Hash_table::id_type id); // locate

  size_t size() const { return table.size(); }
  void print(std::ostream&) const;
};

class global_identifier : public identifier
{ const shared_share address;
public:
  explicit global_identifier(Hash_table::id_type id);
  virtual ~global_identifier()  {}
  virtual void evaluate(level l) const;
};

class local_identifier : public identifier
{ size_t depth, offset;
public:
  explicit local_identifier(Hash_table::id_type id, size_t i, size_t j)
     : identifier(id), depth(i), offset(j) {}
  virtual void evaluate(level l) const; // only this method is redefined
};

struct overload_data
{ shared_value val;  func_type_p type;
  overload_data() : val(),type(NULL) {}
};

class overload_table
{
public:
  typedef std::vector<overload_data> variant_list;
  typedef std::map<Hash_table::id_type,variant_list> map_type;
private:
  map_type table;
  overload_table(const Id_table&); // copying forbidden
  overload_table& operator=(const Id_table&); // assignment forbidden
public:
  overload_table() : table() {} // the default and only accessible constructor
  ~overload_table(); // destructor of all values referenced in the table
 // accessors
  const variant_list& variants(Hash_table::id_type id) const;
  size_t size() const { return table.size(); }
  void print(std::ostream&) const;
 // manipulators
  void add(Hash_table::id_type id, shared_value v, type_ptr t);
   // insertion
  bool remove(Hash_table::id_type id, const type_expr& arg_t); //deletion
};

struct call_expression : public expression_base
{ expression function, argument;

  call_expression(expression_ptr f,expression_ptr a)
   : function(f.release()),argument(a.release()) {}
  virtual ~call_expression()  { delete function; delete argument; }
  virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const;
};

typedef void (* wrapper_function)(expression_base::level);

struct builtin_value : public value_base
{ wrapper_function val;
  std::string print_name;

  builtin_value(wrapper_function v,const std::string& n)
  : val(v), print_name(n)  {}
  virtual void print(std::ostream& out) const
  { out << '{' << print_name << '}'; }
  builtin_value* clone() const { return new builtin_value(*this); }
  static const char* name() { return "built-in function"; }
private:
  builtin_value(const builtin_value& v)
  : val(v.val), print_name(v.print_name)
  {}
};

struct overloaded_builtin_call : public expression_base
{ wrapper_function f;
  std::string print_name;
  expression argument;

  overloaded_builtin_call(wrapper_function v,const char* n,expression_ptr a)
  : f(v), print_name(n), argument(a.release()) {}
  virtual ~overloaded_builtin_call()  { delete argument; }
  virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const;
};

struct generic_builtin_call : public overloaded_builtin_call
{ typedef overloaded_builtin_call base;

  generic_builtin_call(wrapper_function v,const char* n,expression_ptr a)
  : base(v,n,a) {}
  virtual void evaluate(level l) const;
};

struct id_pat_deleter
{ void operator()(id_pat* p) { destroy_id_pat(p); delete p; }};
typedef std::tr1::shared_ptr<id_pat> shared_pattern;
  

struct lambda_expression : public expression_base
{ const shared_pattern param;
  shared_expression body;

  lambda_expression(const id_pat& p, expression_ptr b);
  virtual ~lambda_expression() {} // subobjects do all the work
  virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const;
};

struct closure_value : public value_base
{ context_ptr cont;
  shared_pattern param;
   // used in evaluation only to count arguments
  shared_expression body;

  closure_value(context_ptr c,
                  const shared_pattern& p,
                  const shared_expression& b) : cont(c), param(p), body(b) {}
  void print(std::ostream& out) const;
  closure_value* clone() const 
  { return new closure_value(cont,param,body); }
  static const char* name() { return "closure"; }
};
typedef std::auto_ptr<closure_value> closure_ptr;
typedef std::tr1::shared_ptr<closure_value> shared_closure;

struct overloaded_closure_call : public expression_base
{ shared_closure fun;
  std::string print_name;
  expression argument;

  overloaded_closure_call
   (shared_closure f,const std::string& n,expression_ptr a)
  : fun(f), print_name(n), argument(a.release()) {}
  virtual ~overloaded_closure_call()  { delete argument; }
  virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const;
};

struct conditional_expression : public expression_base
{ expression condition, then_branch, else_branch;

  conditional_expression(expression_ptr c,expression_ptr t, expression_ptr e)
   : condition(c.release()),then_branch(t.release()), else_branch(e.release())
  {}
  virtual ~conditional_expression()
  { delete condition; delete then_branch; delete else_branch; }
  virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const;
};

struct while_expression : public expression_base
{ expression condition, body, next_part;

  while_expression(expression_ptr c,expression_ptr b, expression_ptr n)
   : condition(c.release()),body(b.release()), next_part(n.release())
  {}
  virtual ~while_expression() 
  { delete condition; delete body; delete next_part; }
  virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const;
};

struct for_expression : public expression_base
{ id_pat pattern; expression in_part, body; subscr_base::sub_type kind;

  for_expression
   (id_pat p, expression_ptr i, expression_ptr b, subscr_base::sub_type k);
  virtual ~for_expression() 
  { destroy_id_pat(&pattern); delete in_part; delete body; }
  virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const;
};

struct inc_for_expression : public expression_base
{ expression count, bound, body; Hash_table::id_type id;

  inc_for_expression
   (Hash_table::id_type i, expression_ptr cnt, expression_ptr bnd,
    expression_ptr b)
  : count(cnt.release()),bound(bnd.release()),body(b.release()),id(i)
  {}
  virtual ~inc_for_expression() 
  { delete count; delete bound; delete body; }
  virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const;
};

struct dec_for_expression : public expression_base
{ expression count, bound, body; Hash_table::id_type id;

  dec_for_expression
   (Hash_table::id_type i, expression_ptr cnt, expression_ptr bnd,
    expression_ptr b)
  : count(cnt.release()),bound(bnd.release()),body(b.release()),id(i)
  {}
  virtual ~dec_for_expression() 
  { delete count; delete bound; delete body; }
  virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const;
};

struct assignment_expr : public expression_base
{ Hash_table::id_type lhs;
  expression rhs;

  assignment_expr(Hash_table::id_type l,expression_ptr r)
   : lhs(l),rhs(r.release()) {}
  virtual ~assignment_expr() { delete rhs; }
  virtual void print(std::ostream& out) const;
};

class global_assignment : public assignment_expr
{ shared_share address;
public:
  global_assignment(Hash_table::id_type l,expression_ptr r);
  virtual ~global_assignment() {}
  virtual void evaluate(level l) const;
};

class local_assignment : public assignment_expr
{ size_t depth, offset;
public:
  local_assignment(Hash_table::id_type l, size_t i,size_t j, expression_ptr r);
  virtual ~local_assignment() {}
  virtual void evaluate(level l) const;
};

struct component_assignment : public assignment_expr
{ expression index;
  component_assignment
   (Hash_table::id_type a,expression_ptr i,expression_ptr r)
   : assignment_expr(a,r), index(i.release()) {}

  virtual ~component_assignment() { delete index; }
  virtual void print (std::ostream& out) const;

  void assign(level l,shared_value& aggregate,subscr_base::sub_type kind) const;
};

class global_component_assignment : public component_assignment
{ subscr_base::sub_type kind;
  shared_share address;
public:
  global_component_assignment
    (Hash_table::id_type a,expression_ptr i,expression_ptr r,
     subscr_base::sub_type k);
  virtual void evaluate(level l) const;
};

class local_component_assignment : public component_assignment
{ subscr_base::sub_type kind;
  size_t depth, offset;
public:
  local_component_assignment 
   (Hash_table::id_type l, expression_ptr i,size_t d, size_t o,
    expression_ptr r, subscr_base::sub_type k);
  virtual void evaluate(level l) const;
};

struct seq_expression : public expression_base
{ expression first,last;

  seq_expression(expression_ptr f,expression_ptr l)
   : first(f.release()),last(l.release()) {}
  virtual ~seq_expression()  { delete first; delete last; }
  virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const;
};

struct vector_value : public value_base
{ matrix::Vector<int> val;

  explicit vector_value(const std::vector<int>& v) : val(v)  {}
  ~vector_value() {}
  virtual void print(std::ostream& out) const;
  vector_value* clone() const { return new vector_value(*this); }
  static const char* name() { return "vector"; }
private:
  vector_value(const vector_value& v) : val(v.val) {}
};

typedef std::auto_ptr<vector_value> vector_ptr;
typedef std::tr1::shared_ptr<vector_value> shared_vector;

struct matrix_value : public value_base
{ matrix::Matrix<int> val;

  explicit matrix_value(const matrix::Matrix<int>& v) : val(v)  {}
  ~matrix_value() {}
  virtual void print(std::ostream& out) const;
  matrix_value* clone() const { return new matrix_value(*this); }
  static const char* name() { return "matrix"; }
private:
  matrix_value(const matrix_value& v) : val(v.val) {}
};

typedef std::auto_ptr<matrix_value> matrix_ptr;
typedef std::tr1::shared_ptr<matrix_value> shared_matrix;

struct rational_vector_value : public value_base
{ latticetypes::RatWeight val;

  explicit rational_vector_value(const latticetypes::RatWeight& v):val(v){}
  rational_vector_value(const matrix::Vector<int>& v,int d)
   : val(v,d)  { val.normalize(); }
  ~rational_vector_value() {}
  virtual void print(std::ostream& out) const;
  rational_vector_value* clone() const
   { return new rational_vector_value(*this); }
  static const char* name() { return "rational vector"; }
private:
  rational_vector_value(const rational_vector_value& v) : val(v.val) {}
};

typedef std::auto_ptr<rational_vector_value> rational_vector_ptr;
typedef std::tr1::shared_ptr<rational_vector_value> shared_rational_vector;


extern const type_expr rat_type; // \.{rat}
extern const type_expr str_type; // \.{string}
extern const type_expr vec_type; // \.{vec}
extern const type_expr ratvec_type; // \.{ratvec}
extern const type_expr mat_type; // \.{mat}
extern const type_expr row_of_int_type; // \.{[int]}
extern const type_expr row_of_rat_type; // \.{[rat]}
extern const type_expr row_of_vec_type; // \.{[vec]}
extern const type_expr row_row_of_int_type; // \.{[[int]]}
extern const type_expr pair_type; // \.{(*,*)}
extern const type_expr int_int_type; 

extern Id_table* global_id_table;

extern overload_table* global_overload_table;

extern std::ostream* output_stream;

extern int verbosity;


void initialise_evaluator();

void reset_evaluator ();

expression convert_expr(const expr& e, type_expr& type)
  throw(std::bad_alloc,program_error);

std::ostream& operator<< (std::ostream& out, const Id_table& p);

std::ostream& operator<< (std::ostream& out, const overload_table& p);

expression resolve_overload
  (const expr& e,
   type_expr& type,
   const overload_table::variant_list& variants);

type_ptr analyse_types(const expr& e,expression_ptr& p)
   throw(std::bad_alloc,std::runtime_error);

void install_function
 (wrapper_function f,const char*name, const char* type_string);

void id_mat_wrapper (expression_base::level);
void transpose_mat_wrapper (expression_base::level);
void transpose_vec_wrapper (expression_base::level);

void vv_prod_wrapper (expression_base::level);
void mv_prod_wrapper (expression_base::level);
void mm_prod_wrapper (expression_base::level);
void vm_prod_wrapper(expression_base::level l);


template <typename T>
  std::string str(T n) { std::ostringstream s; s<<n; return s.str(); }

} }
#endif

