
#include "lexer.h"
using namespace std; 
namespace atlas
{ namespace interpreter
   {

String_pool::String_pool(size_t b)
: block_size(b),start(NULL),point(NULL),chars_left(0),prev(NULL)
{}

char* String_pool::store(const char* s,size_t len)
{ size_t n= len>=block_size ? len+1 : block_size;
            // size of new block if needed
  if (len>=chars_left) // then the current block cannot store |s|, so allocate
  { if (start!=NULL) // initial block needs no backing up
    { String_pool* p=new String_pool(block_size);
      p->prev=prev; p->start=start; // only these fields need backing up
      prev=p; // link to old blocks, needed for destruction
    }
    start=point=new char[n]; // now we can copy to where |point| points
    chars_left=n;
  }
  strncpy(point,s,len);  char* result=point;
  chars_left-=len+1;
  point+=len;  *point++='\0';
  return result;
}

String_pool::~String_pool(void)
{ delete prev; delete[] start; }

Hash_table::Hash_table(size_t init,size_t block_size)
: pool(block_size)
, mod(init<2?2:init), hash_tab(mod), name_tab(0)
{ name_tab.reserve(max_fill()); for(int i=0; i<mod; ++i) hash_tab[i]=-1;
}

Hash_table::id_type Hash_table::hash
        (const char* s, size_t l) const // |s| a string of length |l>0|
{ long r=static_cast<unsigned char>(*s++)%mod;
  for ( --l; l>0 ; --l) r=(static_cast<unsigned char>(*s++)+(r<<8))%mod;
  return static_cast<id_type>(r);
}

Hash_table::id_type Hash_table::do_match
	(const char* str, size_t l, bool copy_string)
{ id_type i,h=hash(str,l);
  while ((i=hash_tab[h])>=0)
    if (std::strncmp(name_tab[i],str,l)==0 && name_tab[i][l]=='\0') return i;
          /* identifier found in table */
    else if (++h==mod) h=0; /* move past occupied slot */
  if (name_tab.size()>=max_fill())
  { 
    { vector<id_type>(mod=2*mod-1).swap(hash_tab); // keep it odd
      for (int h=0; h<mod; ++h) hash_tab[h]=-1;
      for (size_t i=0; i<name_tab.size(); ++i)
      { int h=hash(name_tab[i],strlen(name_tab[i])); // rehash the string
        while (hash_tab[h]>=0)  if (++h==mod) h=0; /* find empty slot */
        hash_tab[h]=i;
      }
      name_tab.reserve(max_fill());
    }
    h=hash(str,l); 
    while (hash_tab[h]>=0) 
      if (++h==mod) h=0; /* find free slot */
  }
  hash_tab[h]= name_tab.size(); /* this number represents the new identifier */
  name_tab.push_back(copy_string ? pool.store(str,l) : str);
  return hash_tab[h];
}

Lexical_analyser::Lexical_analyser
  (BufferedInput& source, Hash_table& hash,
   const char** keywords, const char** type_names)
: input(source),id_table(hash),nesting(0)
 ,prevent_termination('\0'),state(initial)
{ 
  { for (size_t i=0; keywords[i]!=0; ++i)
      id_table.match_literal(keywords[i]);
    keyword_limit=id_table.nr_entries();
    for (size_t i=0; type_names[i]!=0; ++i)
      id_table.match_literal(type_names[i]);
    type_limit=id_table.nr_entries();
  }
  comment_start=comment_end=256; // a non-|char| value
}

bool Lexical_analyser::reset()
{ nesting=0; state=initial; input.reset(); return input.getline(); }

void Lexical_analyser::skip_space(void)
{ if (prevent_termination!='\0') input.push_prompt(prevent_termination);
  do
  { char c=input.shift();
    if (isspace(c))
    { if (c=='\n' && prevent_termination=='\0' && nesting==0) break;
         // newline is not skipped if some command could end here
      continue; // all other space is skipped
    }
    if (c==comment_start)
      if (comment_end=='\n' or comment_end==comment_start)
      { do c=input.shift(); while (c!=comment_end && c!='\0');
        input.unshift(); // reconsider newline character
      }
      else // skip comment, allowing for nested comment groups
      { int level=1;
        input.push_prompt(comment_start);
        do
        {
          do c=input.shift();
          while (c!= comment_start and c!=comment_end and c!='\0');
          if (c=='\0')
            break; // force out of comment loop
          if (c==comment_start)
          { ++level;
            input.push_prompt(comment_start);
          }
          else
          { --level;
            input.pop_prompt();
          } // |c==comment_end|
        }
        while (level>0);
      }
    else break; // non-space and non-comment character
  } while(true);
  if (prevent_termination!='\0') input.pop_prompt();
  input.unshift(); // prepare to re-read character that ended space
}

char* Lexical_analyser::scan_quoted_string()
{ const char* start=input.point(),*end;
  int nr_quotes=0; // number of escaped quotes
  do
  { char c;
    do c=input.shift(); while (c!='"' && c!='\n' && c!='\0');
    if (c!='"')
    { input.unshift(); end=input.point();
      int l0,c0,l1,c1;
      input.locate(start,l0,c0); input.locate(end,l1,c1);
      input.show_range(cerr,l0,c0,l1,c1);
      cerr << "Closing string denotation.\n";
      break;
    }
    else if ((c=input.shift())!='"')
    { input.unshift(); end=input.point()-1; break; }
    else ++nr_quotes; // for doubled quotes, continue
  } while (true);
  size_t len=end-start-nr_quotes;
  char* s=new char[len+1];
  while (start<end) // copy characters, undoubling doubled quotes
    if ((*s++=*start++)=='"') ++start;
  *s='\0'; return s-len;
}

int Lexical_analyser::get_token(YYSTYPE *valp, YYLTYPE* locp)
{ if (state==ended) { state=initial; return 0; } // send end of input
  skip_space(); prevent_termination='\0';
  input.locate(input.point(),locp->first_line,locp->first_column);
  int code; char c=input.shift();
  if (isalpha(c)) 
                  { const char* p=input.point()-1; // start of token
                    do c=input.shift(); while(isalpha(c) || isdigit(c) || c=='_');
                    input.unshift();
                    Hash_table::id_type id_code=id_table.match(p,input.point()-p);
                    if (id_code>=type_limit) { valp->id_code=id_code; code=IDENT; }
                    else if (id_code>=keyword_limit)
                    { valp->type_code=id_code-keyword_limit; code=TYPE; }
                    else
                    { code=QUIT+id_code;
                      switch(code)
                      {
                        case LET: ++nesting; input.push_prompt('L'); break;
                        case IF:
                        case WHILE:
                        case FOR: ++nesting; input.push_prompt('G'); break;
                        case IN: if (input.top_prompt()=='L')
                        { --nesting; input.pop_prompt(); prevent_termination='I'; }
                        break;
                        case FI:
                        case OD: --nesting; input.pop_prompt(); break;
                        case WHATTYPE: prevent_termination='W'; break;
                      }
                    }
                  }
  else if (isdigit(c)) 
                       { const char* p=input.point()-1; // start of token
                         do c=input.shift(); while(isdigit(c));
                         input.unshift();
                         const char* end=input.point();
                         unsigned long val=*p++-'0';   while (p<end) val=10*val+(*p++-'0');
                         valp->val=val; code=INT;
                       }
  else 
       { switch(c)
         {      case '"': 
                       { valp->expression=make_string_denotation(scan_quoted_string());
                         code=STRING;
                       }
        
         break; case '(':
                case '{':
                case '[': ++nesting; input.push_prompt(c); code=c;
         break; case ')':
                case '}':
                case ']': --nesting; input.pop_prompt(); code=c;
         break; case '<':
                case '>':
                if (state==initial)
                { code= c=='<' ? FROMFILE :
                        input.shift()=='>' ? ADDTOFILE :
                        (input.unshift(),TOFILE) ;
       	   
       	   if ((skip_space(),c=input.shift())=='"')
       	   { char* s=scan_quoted_string(); file_name=s; delete[] s; }
       	   else
       	   { file_name="";
       	       while (!isspace(c)){ file_name+=c; c=input.shift(); }
       	       input.unshift();
       	   }
         break;
                }
                prevent_termination=c;
                code = OPERATOR; valp->oper.priority=2;
                if (input.shift()=='=')
                  valp->oper.id=id_table.match_literal(c=='<' ? "<=" : ">=");
                else
                  { input.unshift();
                      valp->oper.id=id_table.match_literal(c=='<' ? "<" : ">");
                  }
         break; case ':': prevent_termination=c;
           code = input.shift()=='=' ? BECOMES  : (input.unshift(),c);
         break; case '=':
                valp->oper.id = id_table.match_literal("=");
                valp->oper.priority = 2; // in case
       	 prevent_termination=code=c;
         break; case '!':
                if (input.shift()=='=')
                { code = OPERATOR;
                  valp->oper.id = id_table.match_literal("!=");
                  valp->oper.priority = 2;
                  prevent_termination='=';
                }
                else
                  code= (input.unshift(),c);
                // currently unused; might some day be factorial operator
         break; 
                    case '+': prevent_termination=code=c;
                       code = OPERATOR;
                       valp->oper.id = id_table.match_literal("+");
                       valp->oper.priority = 4;
                break; case '-': prevent_termination=c;
                       if (input.shift()=='>')
                          code= ARROW;
                       else
                       { input.unshift();
                         code = OPERATOR;
                         valp->oper.id = id_table.match_literal("-");
                         valp->oper.priority = 4;
                       }
                break; case '*': case '%': case '/': prevent_termination=c;
                       code = OPERATOR;
                       valp->oper.id =
                          id_table.match_literal(c=='*' ? "*" : c=='%' ? "%" : "/");
                       valp->oper.priority = 6;
                break; case '\\':
                       code = OPERATOR;
                       valp->oper.priority = 6;
                       if (input.shift()=='%')
                       { prevent_termination='%';
                         valp->oper.id = id_table.match_literal("\\%");
                       }
                       else
                       { input.unshift();
                         valp->oper.id = id_table.match_literal("\\");
                       }
                break; case '^': prevent_termination=c;
                       code = OPERATOR;
                       valp->oper.id = id_table.match_literal("^");
                       valp->oper.priority = 7; // exponentiation is right-associative
                break; case '#': prevent_termination=c;
                       code = OPERATOR;
                       valp->oper.id = id_table.match_literal("#");
                       valp->oper.priority = 8; // basic meaning is non-associative
                break;
                case '\n': state=ended; // and {\bf fall through}.
                default: code=c;
         }
       }
  input.locate(input.point(),locp->last_line,locp->last_column);
if (state==initial) state=normal; return code;
}


extern "C"
char* id_completion_func(const char* text, int state)
{ static size_t l; static Hash_table::id_type i,n;
  if (state==0)
  { i=0; n=main_hash_table->nr_entries();
    l=std::strlen(text); // fix length for during search
  }
  while (i<n)
    // |i| is initialised above when |state==0|, and incremented below
  { const char* s=main_hash_table->name_of(i++);
      // get stored identifier and increment loop
    if (std::strncmp(text,s,l) == 0) // is |text| a prefix of |s|?
    { char* res=static_cast<char*>(std::malloc(std::strlen(s)+1));
        // we need a |malloc|ed string
      if (res==NULL)
	return NULL; // if memory gets full here, that's just too bad
      std::strcpy(res,s);
      return res;
    }
  }
  return NULL; /* if loop terminates, report failure */
}


Hash_table* main_hash_table=NULL;

Lexical_analyser* lex=NULL;

   }
}

