
#ifndef LEXER_H
#define LEXER_H

#include  "buffer.h"
#include <vector>
#include  <cstring>
#include  <cstdlib>
#include  "parsetree.h"
#include  "parser.tab.h"
#include <cctype>

namespace atlas { namespace interpreter {


class String_pool
{ public:
    static const size_t default_block_size=512;
    String_pool(size_t b=default_block_size); // create pool and set block size
    char* store(const char* s, size_t length); // store string of given length
    ~String_pool();
  private:
    String_pool(const String_pool&); // copying forbidden
    String_pool& operator=(const String_pool&); // assignment forbidden

    size_t block_size; // block size of the pool
    char* start; // start of block in current usage
    char* point; // first character available
    size_t chars_left; // number of characters left in current block
    String_pool* prev; // start of list of previous blocks, for deallocation
};

class Hash_table
{ public: typedef short id_type; // type of value representing identifiers
private: // data members
    String_pool pool;
    id_type mod;  // hash modulus, the number of slots present
    std::vector<id_type> hash_tab;
    std::vector<const char*> name_tab;
public: // interface
    Hash_table(size_t init=initial_hash_mod
              ,size_t block_size=String_pool::default_block_size);
    id_type match(const char* s, size_t l) { return do_match(s,l,true); }
    id_type match_literal(const char* s) /* literal null-terminated string */
      { return do_match(s,strlen(s),false); }
    const char* name_of(id_type nr) const { return name_tab[nr]; }
    id_type nr_entries() const { return name_tab.size(); }
private: // auxiliary functions
    static const id_type initial_hash_mod=97;
    id_type hash(const char* s, size_t l) const; // auxiliary hash function
    id_type do_match(const char* s, size_t l, bool copy_string);
    size_t max_fill() const { return static_cast<id_type>(mod*3L/4); }
                    // maximum of entries in use
};

class Lexical_analyser
{ enum states  { initial, normal, ended };
BufferedInput& input;
  Hash_table& id_table;
  Hash_table::id_type keyword_limit; // first non-keyword identifier
  Hash_table::id_type type_limit; // first non-type identifier
  int nesting; // number of pending opening symbols
  char prevent_termination; // either |'\0'| or character requiring more input
  int comment_start, comment_end; // characters that start/end a comment
  states state; // to trigger special behaviour
  std::string file_name; // stores a file name for I/O redirection
public:
  Lexical_analyser
    (BufferedInput&, Hash_table&, const char**, const char** type_names);
  int get_token(YYSTYPE *valp, YYLTYPE* locp);
  bool reset(); // get clean slate, return |false| if |getline()| fails
  void set_comment_delims (char c, char d) 
          { comment_start=c; comment_end=d; }
  const char* scanned_file_name() const { return file_name.c_str(); }
private:
  void skip_space();
  char* scan_quoted_string();
};


extern "C"
char* id_completion_func(const char* text, int state);


extern Hash_table* main_hash_table;

extern Lexical_analyser* lex;


} }
#endif

