#include  "parsetree.h"
#include  "parser.tab.h"
#include  <iostream>
#include  <fstream>
#include  <readline/readline.h>
#include  <readline/history.h>
#include  "buffer.h"
#include  "lexer.h"
#include  "built-in-types.h"
#include  "constants.h"
#include  <stdexcept>
#include  "evaluator.h"
#include  <cstring>


extern "C"
{ int yylex (YYSTYPE *, YYLTYPE *);
int yyparse( expr* parsed_expr, int* verbosity );
}

namespace { 
            const char* keywords[] =
             {"quit"
             ,"set","let","in"
             ,"if","then","else","elif","fi"
             ,"and","or","not"
             ,"while","do","od","next","for","from","downto"
             ,"true","false"
             ,"quiet","verbose"
             ,"whattype","showall"
             ,NULL};
 }

extern "C"
int yylex(YYSTYPE *valp, YYLTYPE *locp)
{ return atlas::interpreter::lex->get_token(valp,locp); }

extern "C"
void yyerror (YYLTYPE* locp, expr* ,int* ,char const *s)
{ atlas::interpreter::main_input_buffer->show_range(std::cerr,
   locp->first_line, locp->first_column,
   locp->last_line,  locp->last_column);
  std::cerr << s << std::endl;
  atlas::interpreter::main_input_buffer->close_includes();
}


int main(int argc, char** argv)
{ using namespace std; using namespace atlas::interpreter;

  
  bool use_readline = argc<2 or std::strcmp(argv[1],"-nr")!=0;
BufferedInput input_buffer("expr> "
                            ,use_readline ? readline : NULL
			    ,use_readline ? add_history : NULL);
  main_input_buffer= &input_buffer;
Hash_table hash; main_hash_table= &hash;
Lexical_analyser ana(input_buffer,hash,keywords,prim_names); lex=&ana;
Id_table main_table;  global_id_table=&main_table;
overload_table main_overload_table;
  global_overload_table=&main_overload_table;

  
    using_history();
    rl_completion_entry_function = id_completion_func; // set up input completion
  
  ana.set_comment_delims('{','}');
  initialise_evaluator(); initialise_builtin_types();

  cout << "Enter expressions:\n";
  while (ana.reset()) // get a fresh line for lexical analyser, or quit
  { expr parse_tree;
    int old_verbosity=verbosity;
    ofstream redirect; // if opened, this will be closed at end of loop
    if (yyparse(&parse_tree,&verbosity)!=0)
      continue; // syntax error or non-expression
    if (verbosity!=0) // then some special action was requested
    { if (verbosity<0) break; // \.{quit} command
      if (verbosity==2 or verbosity==3)
        // indicates output redirection was requested
      { 
        { redirect.open(ana.scanned_file_name() ,ios_base::out |
             (verbosity==2 ? ios_base::trunc : ios_base::app));
          if (redirect.is_open()) output_stream = &redirect;
          else
          { cerr << "Failed to open " << ana.scanned_file_name() << endl;
            continue;
          }
        }
        verbosity=old_verbosity; // verbosity change was temporary
      }
      if (verbosity==1) //
        cout << "Expression before type analysis: " << parse_tree << endl;
    }
    
    { bool type_OK=false;
      try
      { expression_ptr e;
        type_ptr type=analyse_types(parse_tree,e);
        type_OK=true;
        if (verbosity>0)
          *output_stream << "Type found: " << *type << endl 
    	<< "Converted expression: " << *e << endl;
        e->evaluate(expression_base::single_value);
        shared_value v=pop_value();
        static type_expr empty(type_list_ptr(NULL));
        if (*type!=empty) *output_stream << "Value: " << *v << endl;
        destroy_expr(parse_tree);
      }
      catch (runtime_error& err)
      { if (type_OK)
          cerr << "Runtime error:\n  " << err.what() << "\nEvaluation aborted.";
        else cerr << err.what();
        cerr << std::endl;
        reset_evaluator(); main_input_buffer->close_includes();
      }
      catch (logic_error& err)
      { cerr << "Internal error: " << err.what() << ", evaluation aborted.\n";
        reset_evaluator(); main_input_buffer->close_includes();
      }
      catch (exception& err)
      { cerr << err.what() << ", evaluation aborted.\n";
        reset_evaluator(); main_input_buffer->close_includes();
      }
    }
    output_stream= &cout; // reset output stream if it was changed
  }
  clear_history();
  // clean up (presumably disposes of the lines stored in history)
  cout << "Bye.\n";
}


