#include  "parsetree.h"
#include  "lexer.h"


const expr_list null_expr_list=NULL;

struct partial_formula dummy_formula;

namespace atlas
{ namespace interpreter
  {
extern "C" {

void destroy_expr(expr e)
{ switch (e.kind)
  { 
    case integer_denotation: case boolean_denotation: break;
    case string_denotation: delete[] e.e.str_denotation_variant; break;
    
    case applied_identifier: break;
    
    case list_display: destroy_exprlist(e.e.sublist);
    break;
    
    case tuple_display: destroy_exprlist(e.e.sublist);
    break;
    
    case function_call:
      destroy_expr(e.e.call_variant->fun);
      destroy_expr(e.e.call_variant->arg);
      delete e.e.call_variant;
    break;
    
    case let_expr:
    { let lexp=e.e.let_variant;
      destroy_id_pat(&lexp->pattern);
      destroy_expr(lexp->val);
      destroy_expr(lexp->body);
      delete lexp;
    }
    break;
    
    case lambda_expr:
    { lambda fun=e.e.lambda_variant;
      destroy_id_pat(&fun->pattern);
      destroy_type(fun->arg_type);
      destroy_expr(fun->body);
    }
    break;
    
    case conditional_expr:
      destroy_expr(e.e.if_variant->condition);
      destroy_expr(e.e.if_variant->then_branch);
      destroy_expr(e.e.if_variant->else_branch);
      delete e.e.if_variant;
    break;
    
    case while_expr:
      destroy_expr(e.e.while_variant->condition);
      destroy_expr(e.e.while_variant->body);
      destroy_expr(e.e.while_variant->next_part);
      delete e.e.while_variant;
    break;
    case for_expr:
      destroy_id_pat(&e.e.for_variant->id);
      destroy_expr(e.e.for_variant->in_part);
      destroy_expr(e.e.for_variant->body);
      delete e.e.for_variant;
    break;
    case cfor_expr:
      destroy_expr(e.e.cfor_variant->count);
      destroy_expr(e.e.cfor_variant->bound);
      destroy_expr(e.e.cfor_variant->body);
      delete e.e.cfor_variant;
    break;
    
    case subscription:
      destroy_expr(e.e.subscription_variant->array);
      destroy_expr(e.e.subscription_variant->index);
      delete e.e.subscription_variant;
    break;
    
    case cast_expr:
      destroy_expr(e.e.cast_variant->exp); delete e.e.cast_variant;
    break;
    
    case op_cast_expr:
      delete e.e.op_cast_variant;
    break;
    
    case ass_stat:
      destroy_expr(e.e.assign_variant->rhs); delete e.e.assign_variant;
    break;
    
    case comp_ass_stat:
      destroy_expr(e.e.comp_assign_variant->index);
      destroy_expr(e.e.comp_assign_variant->rhs);
      delete e.e.comp_assign_variant;
    break;
    
    case seq_expr:
      destroy_expr(e.e.sequence_variant->first);
      destroy_expr(e.e.sequence_variant->last);
      delete e.e.sequence_variant;
    break;
  }
}

expr make_int_denotation (int val)
{ expr result; result.kind=integer_denotation;
result.e.int_denotation_variant=val; return result;
}
expr make_string_denotation(char* val)
{ expr result; result.kind=string_denotation;
result.e.str_denotation_variant=val; return result;
}
expr make_bool_denotation(int val)
{ expr result; result.kind=boolean_denotation;
result.e.int_denotation_variant=val; return result;
}

expr make_applied_identifier (id_type id)
{ expr result; result.kind=applied_identifier;
  result.e.identifier_variant=id; return result;
}

void destroy_exprlist(expr_list l)
{ while (l!=NULL)
  { destroy_expr(l->e); expr_list this_node=l; l=l->next; delete this_node; }
}

expr_list make_exprlist_node(expr e, expr_list l)
{ expr_list n=new exprlist_node; n->e=e; n->next=l; return n; }
expr_list reverse_expr_list(expr_list l)
{ expr_list r=null_expr_list;
  while (l!=null_expr_list) { expr_list t=l; l=t->next; t->next=r; r=t; }
  return r;
}
expr wrap_list_display(expr_list l)
{ expr result; result.kind=list_display; result.e.sublist=l; return result;
}

expr wrap_tuple_display(expr_list l)
{ expr result; result.kind=tuple_display; result.e.sublist=l; return result;
}

expr make_application_node(expr f, expr_list args)
{ app a=new application_node; a->fun=f;
  if (args!=NULL && args->next==NULL) // a single argument
  { a->arg=args->e; delete args; }
  else a->arg=wrap_tuple_display(args);
  expr result; result.kind=function_call; result.e.call_variant=a;
  return result;
}

expr make_unary_call(id_type name, expr arg)
{ app a=new application_node; a->fun=make_applied_identifier(name); a->arg=arg;
  expr result; result.kind=function_call; result.e.call_variant=a;
  return result;
}

expr make_binary_call(id_type name, expr x, expr y)
{ expr_list args=make_exprlist_node(x,make_exprlist_node(y,NULL));
  return make_application_node(make_applied_identifier(name),args);
}

form_stack start_formula (expr e, Hash_table::id_type op, int prio)
{ form_stack result = new partial_formula;
  result->left_subtree=e; result->op=op; result->prio=prio;
  result->prev=NULL;
  return result;
}

form_stack start_unary_formula (id_type op, int prio)
{ form_stack result = new partial_formula;
  result->op=op; result->prio=prio;
  result->prev=unary_marker;
  return result;
}

form_stack extend_formula (form_stack pre, expr e,id_type op, int prio)
{ form_stack result=NULL;
  while (pre!=NULL and (pre->prio>prio or pre->prio==prio and prio%2==0))
  { delete result; result=pre; // clean up, but holding on to one node
    
    if (pre->prev!=unary_marker)
    { e = make_binary_call(pre->op,pre->left_subtree,e);
      pre=pre->prev;
    }
    else
    { e = make_unary_call(pre->op,e);
      pre=NULL;
    } 
  }

  if (result==NULL)
    result=new partial_formula; // allocate if no nodes were combined
  result->left_subtree=e; result->op=op; result->prio=prio;
  result->prev=pre; return result;
}

expr end_formula (form_stack pre, expr e)
{ while (pre!=NULL)
  { form_stack t=pre;
    
    if (pre->prev!=unary_marker)
    { e = make_binary_call(pre->op,pre->left_subtree,e);
      pre=pre->prev;
    }
    else
    { e = make_unary_call(pre->op,e);
      pre=NULL;
    } 
    delete t;
  }
  return e;
}

void destroy_formula(form_stack s)
{ while (s!=NULL and s!=unary_marker)
  { form_stack t=s; s=s->prev; delete t;
  }
}

patlist make_pattern_node(patlist next,struct id_pat* body)
{ patlist l=new pattern_node; l->next=next; l->body=*body; return l; }

void destroy_pattern(patlist p)
{ while (p!=NULL)
  { patlist q=p; p=p->next; destroy_id_pat(&q->body); delete q; }
}

void destroy_id_pat(struct id_pat* p)
{ if ((p->kind & 0x2)!=0)
    destroy_pattern(p->sublist);
}

patlist reverse_patlist(patlist p)
{ patlist q=NULL;
  while (p!=NULL)
  { patlist t=q; q=p; p=q->next; q->next=t; }
  return q;
}

void destroy_letlist(let_list l)
{ while (l!=NULL)
    { let_list p=l; l=l->next;
      destroy_id_pat(&p->pattern);
      destroy_expr(p->val);
      delete p;
    }
}

let_list add_let_node(let_list prev,struct id_pat pattern, expr val)
{ let_list l=new let_node;
  l->pattern=pattern; l->val=val; l->next=prev;
  return l;
}

expr make_let_expr_node(let_list decls, expr body)
{ let l=new let_expr_node; l->body=body;
  expr result; result.kind=let_expr; result.e.let_variant=l;
  try
  { if (decls->next==NULL) // single declaration
    { l->pattern=decls->pattern; l->val=decls->val;
      delete decls;
    }
    else
    { l->pattern.kind=0x2; l->pattern.sublist=NULL;
      l->val=wrap_tuple_display(null_expr_list);
      while (decls!=NULL)
      { l->pattern.sublist =
	  make_pattern_node(l->pattern.sublist,&decls->pattern);
	l->val.e.sublist=make_exprlist_node(decls->val,l->val.e.sublist);
	let_list p=decls; decls=p->next; delete p;
      }
    }
    return result;
  }
  catch(...)
  { destroy_expr(result);
    throw;
  }
}

expr make_lambda_node(patlist patl, ptr typel, expr body)
{ lambda fun=new lambda_node; fun->body=body;
  if (patl!=NULL and patl->next==NULL)
  { fun->pattern=patl->body; delete patl; // clean up node
    fun->arg_type = first_type(typel);
  }
  else
  { fun->pattern.kind=0x2; fun->pattern.sublist=patl;
    fun->arg_type=mk_tuple_type(typel);
  }
  expr result; result.kind=lambda_expr; result.e.lambda_variant=fun;
  return result;
}

expr make_conditional_node(expr c, expr t, expr e)
{ cond n=new conditional_node; n->condition=c;
  n->then_branch=t; n->else_branch=e;
expr result; result.kind=conditional_expr; result.e.if_variant=n;
  return result;
}

expr make_while_node(expr c, expr b, expr n)
{ w_loop w=new while_node; w->condition=c; w->body=b; w->next_part=n;
expr result; result.kind=while_expr; result.e.while_variant=w;
  return result;
}
expr make_for_node(struct id_pat id, expr ip, expr b)
{ f_loop f=new for_node; f->id=id; f->in_part=ip; f->body=b;
expr result; result.kind=for_expr; result.e.for_variant=f;
  return result;
}
expr make_cfor_node(id_type id, expr count, expr bound, bool up, expr b)
{ c_loop c=new cfor_node; c->id=id; c->count=count; c->bound=bound;
  c->up=up; c->body=b;
expr result; result.kind=cfor_expr; result.e.cfor_variant=c;
  return result;
}

expr make_subscription_node(expr a, expr i)
{ sub s=new subscription_node; s->array=a; s->index=i;
  expr result; result.kind=subscription; result.e.subscription_variant=s;
  return result;
}

expr make_cast(ptr type, expr exp)
{ cast c=new cast_node; c->type=type; c->exp=exp;
 expr result; result.kind=cast_expr; result.e.cast_variant=c;
   return result;
}

expr make_op_cast(id_type name,ptr type)
{ op_cast c=new op_cast_node; c->oper=name; c->type=type;
 expr result; result.kind=op_cast_expr; result.e.op_cast_variant=c;
   return result;
}

expr make_assignment(id_type lhs, expr rhs)
{ assignment a=new assignment_node; a->lhs=lhs; a->rhs=rhs;
 expr result; result.kind=ass_stat; result.e.assign_variant=a;
   return result;
}

expr make_comp_ass(expr lhs, expr rhs)
{ comp_assignment a=new comp_assignment_node;
a->aggr=lhs.e.subscription_variant->array.e.identifier_variant;
  a->index=lhs.e.subscription_variant->index;
  delete lhs.e.subscription_variant;
  a->rhs=rhs;
expr result; result.kind=comp_ass_stat; result.e.comp_assign_variant=a;
  return result;
}

expr make_sequence(expr first, expr last)
{ sequence s=new sequence_node; s->first=first; s->last=last;
 expr result; result.kind=seq_expr; result.e.sequence_variant=s;
   return result;
}

id_type lookup_identifier(const char* name)
{ return main_hash_table->match_literal(name); }

void include_file()
{ main_input_buffer->push_file(lex->scanned_file_name()); }

}

std::ostream& operator<< (std::ostream& out, expr e)
{ switch (e.kind)
  { 
    case integer_denotation: out << e.e.int_denotation_variant; break;
    case string_denotation:
      out << '"' << e.e.str_denotation_variant << '"'; break;
    case boolean_denotation:
      out << (e.e.int_denotation_variant!=0 ? "true" : "false"); break;
    
    case applied_identifier:
      out << main_hash_table->name_of(e.e.identifier_variant);
    break;
    
    case list_display:
    { expr_list l=e.e.sublist;
      if (l==NULL) out << "[]";
      else
      { out << '[';
        do { out << l->e; l=l->next; out << (l==NULL ? ']' : ',');}
        while (l!=NULL);
      }
    }
    break;
    
    case tuple_display:
    { expr_list l=e.e.sublist;
      if (l==NULL) out << "()";
      else
      { out << '(';
        do { out << l->e; l=l->next; out << (l==NULL ? ')' : ',');}
        while (l!=NULL);
      }
    }
    break;
    
    case function_call:
    { app a=e.e.call_variant;
      expr fun=a->fun,arg=a->arg;
      if (fun.kind==applied_identifier) out << fun;
      else out << '(' << fun << ')';
      if (arg.kind==tuple_display) out << arg;
      else out << '(' << arg << ')';
    }
    break;
    
    case let_expr:
    { let lexp=e.e.let_variant;
      out << "let " << lexp->pattern << '=' << lexp->val <<	 " in " << lexp->body;
    }
    break;
    
    case lambda_expr:
    { lambda fun=e.e.lambda_variant;
      if ((fun->pattern.kind&0x1)!=0)
        out << '(' << fun->pattern << ')';
      else
        out << fun->pattern;
      out << ':' << fun->body;
    }
    break;
    
    case conditional_expr:
    { cond c=e.e.if_variant; out << " if " << c->condition 
      << " then " << c->then_branch << " else " << c->else_branch << " fi ";
    }
    break;
    
    case while_expr:
    { w_loop w=e.e.while_variant;
      out << " while " << w->condition << " do " << w->body;
      if (w->next_part.kind!=tuple_display or w->next_part.e.sublist!=NULL)
        out << " next " << w->next_part;
      out << " od ";
    }
    break;
    case for_expr:
    { f_loop f=e.e.for_variant;
      out << " for " << f->id.sublist->next->body;
      if (f->id.sublist->body.kind==0x1)
        out << '@' << f->id.sublist->body;
      out << " in " << f->in_part << " do " << f->body << " od ";
    }
    break;
    case cfor_expr:
    { c_loop c=e.e.cfor_variant;
      out << " for " << main_hash_table->name_of(c->id) << " = " << c->count;
      if (c->bound.kind!=tuple_display or c->bound.e.sublist!=NULL)
        out << (c->up!=0 ? " from " : " downto ") << c->bound;
      out << " do " << c->body << " od ";
    }
    break;
    
    case subscription:
    { sub s=e.e.subscription_variant; out << s->array << '[';
      expr i=s->index;
      if (i.kind!=tuple_display) out << i;
      else
      { expr_list l=i.e.sublist;
        if (l!=NULL)
        { out << l->e;
          while ((l=l->next)!=NULL) out << ',' << l->e;
        }
      }
      out << ']';
    }
    break;
    
    case cast_expr:
    { cast c = e.e.cast_variant;
      print_type(out,c->type) << ':' << c->exp ;
    }
    break;
    
    case op_cast_expr:
    { op_cast c = e.e.op_cast_variant;
      print_type(out << main_hash_table->name_of(c->oper) << '@',c->type);
    }
    break;
    
    case ass_stat:
    { assignment ass = e.e.assign_variant;
      out << main_hash_table->name_of(ass->lhs) << ":=" << ass->rhs ;
    }
    break;
    
    case comp_ass_stat:
    { comp_assignment ass = e.e.comp_assign_variant;
      out << main_hash_table->name_of(ass->aggr) << '[' << ass->index << "]:="
          << ass->rhs ;
    }
    break;
    
    case seq_expr:
    { sequence seq = e.e.sequence_variant;
      out << seq->first << ";" << seq->last ;
    }
    break;
  }
  return out;
}

size_t length(expr_list l) 
{ size_t n=0; while(l!=NULL) { ++n; l=l->next;}
  return n;
}

bool is_empty(const expr& e)
{ return e.kind==tuple_display and e.e.sublist==NULL; }

std::ostream& operator<< (std::ostream& out, const id_pat& p)
{ if ((p.kind & 0x1)!=0)
    out << main_hash_table->name_of(p.name);
  if (p.kind==0x3) // both parts present
    out << ':';
  if ((p.kind & 0x2)!=0)
    if (p.sublist==NULL)
      out << "()";
    else
    { out << '(' << p.sublist->body << ',';
      for (patlist l=p.sublist->next; l!=NULL; l=l->next)
      out << l->body << (l->next!=NULL ? ',' : ')');
    }
  return out;
}

  }
}

