
#include "types.h"
#include <cstdlib>

namespace atlas { namespace interpreter {
namespace {

class voiding : public expression_base
{ expression exp;
public:
  voiding(expression_ptr e) : exp(e.release()) {}
  virtual ~voiding(){ delete exp; }
  virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const;
};

} // |namespace|

const char* prim_names[]=
{"int","rat","string","bool",
     "vec", "mat", "ratvec",
     "LieType","RootDatum", "InnerClass", "RealForm", "DualRealForm",
     "CartanClass", 

 "void", NULL };

const type_expr unknown_type; // uses default constructor
 type_expr void_type(type_list_ptr(NULL));
 type_expr int_type(integral_type);
 type_expr bool_type(boolean_type);
const type_expr row_of_type(copy(unknown_type));
const type_expr gen_func_type(copy(unknown_type),copy(unknown_type));

std::vector<shared_value> execution_stack;

std::vector<conversion_record> coerce_table;

namespace {

void voiding::evaluate(level l) const
{ exp->void_eval();
  if (l==single_value)
    wrap_tuple(0);
}

void voiding::print(std::ostream& out) const
{ out << "voided:" << *exp; }

} // |namespace|

type_ptr copy(const type_expr& t) 
{ return type_ptr(new type_expr(t)); }

type_node::type_node(const type_node& n)
 : t(n.t),next(n.next==NULL ? NULL : new type_node(*n.next))
{}

type_node::~type_node()
{ delete next; }

type_list_ptr make_type_singleton(type_ptr t)
{ return type_list_ptr(new type_node(t,type_list_ptr(NULL))); }

type_list_ptr make_type_list(type_ptr t,type_list_ptr l)
{ return type_list_ptr(new type_node(t,l)); }

size_t length(type_list l)
{ size_t len=0;
  for (; l!=NULL; l=l->next) ++len;
  return len;
}

type_expr::type_expr(const type_expr& t) : kind(t.kind)
{ switch (kind)
  { case undetermined_type: break;
    case primitive_type: prim=t.prim; break;
    case row_type:
      component_type=new type_expr(*t.component_type);
    break;
    case tuple_type:
      tuple= t.tuple==NULL ? NULL : new type_node(*t.tuple);
    break;
    case function_type: func=new func_type(*t.func); break;
  }
}

type_expr::~type_expr()
{ switch (kind)
  { case undetermined_type: case primitive_type: break;
    case row_type: delete component_type; break;
    case tuple_type: delete tuple; break;
    case function_type: delete func; break;
  }
}

void type_expr::set_from(type_ptr p)
{ if (kind!=undetermined_type) throw std::logic_error("Illegal set_from");
  kind=p->kind;
  switch(kind) // copy top node
  { case undetermined_type: break;
    case primitive_type: prim=p->prim; break;
    case row_type: component_type=p->component_type; break;
    case function_type: func=p->func; break;
    case tuple_type: tuple=p->tuple; break;
  }
  p->kind=undetermined_type;
  // detach descendants, so auto-pointer destroys top-level only
}

bool type_expr::specialise(const type_expr& pattern)
{ if (pattern.kind==undetermined_type)
    return true; // specialisation to \.* trivially succeeds.
  if (kind==undetermined_type)
    {new(this) type_expr(pattern); return true; }
  if (pattern.kind!=kind) return false; // impossible to refine
  switch(kind)
  { case primitive_type: return prim==pattern.prim;
    case row_type: return component_type->specialise(*pattern.component_type);
    case function_type:
      return func->arg_type.specialise(pattern.func->arg_type) 
         and func->result_type.specialise(pattern.func->result_type);
    case tuple_type:
     
     { type_list l0=tuple, l1=pattern.tuple;
       while (l0!=NULL and l1!=NULL and l0->t.specialise(l1->t))
       { l0=l0->next; l1=l1->next; }
       return l0==NULL and l1==NULL;
       // we succeeded only if both lists terminate simultaneously
     }
    default: return true; // to keep the compiler happy, cannot be reached
  }
}

std::ostream& operator<<(std::ostream& out, type_list l)
{ for (; l!=NULL; l=l->next) out << l->t << ( l->next!=NULL ? "," : "" );
  return out;
}

std::ostream& operator<<(std::ostream& out, const type_expr& t)
{ switch(t.kind)
  { case undetermined_type: out << '*'; break;
    case primitive_type: out << prim_names[t.prim]; break;
    case row_type: out << '[' << *t.component_type << ']'; break;
    case tuple_type:
      out << '(' << t.tuple << ')' ;
    break;
    case function_type:
      out << '(';
      if (t.func->arg_type.kind==tuple_type)
         out << t.func->arg_type.tuple; // naked tuple
      else out << t.func->arg_type; // other component type
      out << "->";
      if (t.func->result_type.kind==tuple_type)
         out << t.func->result_type.tuple; // naked tuple
      else out << t.func->result_type; // other component type
      out << ')'; break;
  }
  return out;
}

bool operator== (const type_expr& x,const type_expr& y)
{ if (x.kind!=y.kind) return false;
  switch (x.kind)
  {  default:
// all cases are listed below, but somehow this keeps \.{g++} from complaining
  case undetermined_type: return true;
    case primitive_type: return x.prim==y.prim;
    case row_type: return *x.component_type==*y.component_type;
    case tuple_type:
    { type_list l0=x.tuple, l1=y.tuple;
      while (l0!=NULL and l1!=NULL and l0->t==l1->t)
	{ l0=l0->next; l1=l1->next; }
      return l0==NULL and l1==NULL;
      // lists must end simultaneously for success
    }
    case function_type:
      return  x.func->arg_type==y.func->arg_type
	 and   x.func->result_type==y.func->result_type;
  }
}

type_ptr make_undetermined_type()
{ return type_ptr(new type_expr); }

type_ptr make_prim_type(primitive_tag p)
{ return p<nr_of_primitive_types ?
    type_ptr(new type_expr(p)) :
    type_ptr(new type_expr(type_list_ptr(NULL)));
}

type_ptr make_row_type(type_ptr c)
{ return type_ptr(new type_expr(c)); }

type_ptr make_tuple_type (type_list_ptr l)
{ return type_ptr(new type_expr(l)); }

type_ptr make_function_type (type_ptr a, type_ptr r)
{ return type_ptr(new type_expr(a,r));
}

extern "C"
ptr mk_type_singleton(ptr t)
{
  return make_type_singleton(type_ptr(static_cast<type_p>(t)))
  .release();
}
extern "C"
ptr mk_type_list(ptr t,ptr l)
{ return make_type_list(type_ptr(static_cast<type_p>(t)),
                        type_list_ptr(static_cast<type_list>(l))).release(); }
extern "C"
ptr mk_prim_type(int p)
{ return make_prim_type(static_cast<primitive_tag>(p)).release(); }
extern "C"
ptr mk_row_type(ptr c)
{ return make_row_type(type_ptr(static_cast<type_p>(c)))
  .release(); }
extern "C"
ptr mk_tuple_type(ptr l)
{ return make_tuple_type(type_list_ptr(static_cast<type_list>(l)))
  .release(); }
extern "C"
ptr mk_function_type(ptr a,ptr r)
{ return make_function_type(type_ptr(static_cast<type_p>(a)),
    type_ptr(static_cast<type_p>(r))).release(); }

extern "C"
void destroy_type(ptr t)
{ delete static_cast<type_p>(t); }
extern "C"
void destroy_type_list(ptr t)
{ delete static_cast<type_list>(t); }

ptr first_type(ptr typel)
{ return &static_cast<type_list>(typel)->t; }

std::ostream& print_type(std::ostream& out, ptr type)
{ return out << *static_cast<type_p>(type); }

type_ptr scan_type(const char*& s);
type_list_ptr scan_type_list(const char*& s);

type_ptr make_type(const char* s)
{ const char* orig=s;
  try { return scan_type(s); }   // provide an lvalue
  catch (std::logic_error e)
  { std::cerr << e.what() << "; original string: '" << orig 
              << "' text remaining: '" << s << "'\n";
    throw;
  // make the error hard to ignore; if thrown probably aborts the program
  }
}

type_ptr scan_type(const char*& s)
{ if (*s=='*') return ++s,make_undetermined_type();
  else if (*s=='[')
    
    { type_ptr c=scan_type(++s);
      if (*s++!=']') throw std::logic_error("Missing ']' in type");
        return make_row_type(c);
    }
  else if (*s=='(')
    
    { type_list_ptr l0=scan_type_list(++s), l1(NULL);
      bool is_tuple=*s==')';
      if (*s=='-' and *++s=='>') l1=scan_type_list(++s);
      if (*s++!=')') throw std::logic_error("Missing ')' in type");
      type_ptr a =
        l0.get()!=NULL and l0->next==NULL ? copy(l0->t) : make_tuple_type(l0);
      if (is_tuple) return a;
      type_ptr r =
        l1.get()!=NULL and l1->next==NULL ? copy(l1->t) : make_tuple_type(l1);
      return make_function_type(a,r);
    }
  else 
       { for (size_t i=0; i<nr_of_primitive_types; ++i)
         { std::string name=prim_names[i];
           if (name.compare(0,name.length(),s,name.length())==0 and
               not isalpha(s[name.length()]))
           { s+=name.length();
             return make_prim_type(static_cast<primitive_tag>(i));
           }
         }
         throw std::logic_error("Type unrecognised");
       }
}


type_list_ptr scan_type_list(const char*& s)
{ if (*s==')' or *s=='-') return type_list_ptr(NULL);
  type_ptr head=scan_type(s);
  if (*s!=',') return make_type_singleton(head);
  return make_type_list(head,scan_type_list(++s));
}

type_ptr unknown_tuple(size_t n)
{ type_list_ptr tl(NULL);
  while (n-->0) tl=make_type_list(copy(unknown_type),tl);
  return make_tuple_type(tl);
}

std::ostream& operator<< (std::ostream& out, const value_base& v)
{ v.print(out); return out; }

void row_value::print(std::ostream& out) const
{ if (val.empty()) out << "[]";
  else
  { out << '[';
    std::vector<shared_value>::const_iterator p=val.begin();
    do { (*p)->print(out); ++p; out << (p==val.end() ? ']' : ','); }
    while (p!=val.end());
  }
}

void tuple_value::print(std::ostream& out) const
{ if (val.empty()) out << "()";
  else
  { out << '(';
    std::vector<shared_value>::const_iterator p=val.begin();
    do { (*p)->print(out); ++p; out << (p==val.end() ? ')' : ','); }
    while (p!=val.end());
  }
}

void push_tuple_components()
{ shared_tuple tuple(get<tuple_value>());
  for (size_t i=0; i<tuple->length(); ++i)
    push_value(tuple->val[i]); // push component
}

void wrap_tuple(size_t n)
{ shared_tuple result(new tuple_value(n));
  while (n-->0) // standard idiom; not |(--n>=0)|, since |n| is unsigned!
    result->val[n]=pop_value();
  push_value(result);
}

shared_value& context::elem(size_t i, size_t j)
{
  context* p=this;
  for (; i-->0; p=p->next.get())
    assert(p->next.use_count()>0 and p->next.get()!=NULL);
  assert(j<p->frame.size());
  return p->frame[j];
}

void conversion::evaluate(level l) const
{ exp->eval();
  (*type.convert)();
  if (l==no_value)
    execution_stack.pop_back();
}

void conversion::print(std::ostream& out) const
{ out << type.name << ':' << *exp; }

void coercion(const type_expr& from,
              const type_expr& to,
              const char* s, conversion_info::conv_f f)
{ coerce_table.push_back(conversion_record(from,to,s,f)); }

bool coerce(const type_expr& from_type, const type_expr& to_type,
	    expression_ptr& e)
{ for (conversion_record*
       it=&coerce_table[0]; it<&*coerce_table.end(); ++it)
    if (from_type==*it->from and to_type==*it->to)
    { e.reset(new conversion(*it,e));
      return true;
    }
  if (to_type==void_type)
  { e.reset(new voiding(e));
     return true;
  }
  return false;
}

conversion_record* row_coercion(const type_expr& final_type,
                                      type_expr& component_type)
{ for (conversion_record* it=&coerce_table[0]; it<&*coerce_table.end(); ++it)
    if (final_type==*it->to and it->from->kind==row_type)
    { component_type.specialise(*it->from->component_type); return it; }
  return NULL;
}

unsigned int is_close (const type_expr& x, const type_expr& y)
{ expression_ptr dummy(NULL);
  if (x.kind==primitive_type or y.kind==primitive_type)
    if (x==y)
      return 0x7;
    else
    { unsigned int flags=0x0;
      if (coerce(x,y,dummy)) flags |= 0x1;
      if (coerce(y,x,dummy)) flags |= 0x2;
      return flags==0 ? flags : flags|0x4;
    }
  if (x.kind!=y.kind)
    return 0x0;
  if (x.kind==row_type)
    return is_close(*x.component_type,*y.component_type);
  if (x.kind!=tuple_type)
    return x==y ? 0x7 : 0x0; // non-aggregate types are only close if equal
  type_list l0=x.tuple, l1=y.tuple; // recurse for tuples, as in |operator==|
  unsigned int flags=0x7;
  while (l0!=NULL and l1!=NULL and (flags&=is_close(l0->t,l1->t))!=0)
    { l0=l0->next; l1=l1->next; }
  return l0==NULL and l1==NULL ? flags : 0x0;
    // lists must end simultaneously for success
}

type_error::type_error(const type_error& e)
 : expr_error(e)
{ type_ptr p=copy(*e.required);
  actual=copy(*e.actual).release(); required=p.release();
}

} }

