
#ifndef TYPES_H
#define TYPES_H


#include <memory> 

#include "parsetree.h"

#include <iostream> // needed for specification of |print| method below
#include <tr1/memory>

#include <vector>
#include <cassert>

#include <stdexcept>

namespace atlas { namespace interpreter {

struct type_expr;
typedef type_expr* type_p;
typedef std::auto_ptr<type_expr> type_ptr;
typedef struct type_node* type_list;
typedef std::auto_ptr<type_node> type_list_ptr;

enum type_tag 
{ undetermined_type, primitive_type, row_type, tuple_type, function_type };

enum primitive_tag
{ integral_type, rational_type, string_type, boolean_type
  , 
    vector_type, matrix_type, rational_vector_type,
    complex_lie_type_type , root_datum_type, inner_class_type, real_form_type,
    dual_real_form_type, Cartan_class_type, 
 
  nr_of_primitive_types };

struct func_type; // must be predeclared for |type_expr|

struct type_expr
{ type_tag kind;
  union
  { primitive_tag prim; // when |kind==primitive|
    type_p component_type; // when |kind==row_type|
    type_list tuple; // when |kind==tuple_type|
    func_type* func; // when |kind==function_type|
  };

  type_expr() : kind(undetermined_type) {}
  explicit type_expr(primitive_tag p)
    : kind(primitive_type) { prim=p; }
  explicit type_expr(type_ptr c)
    : kind(row_type) { component_type=c.release(); }
  explicit type_expr(type_list_ptr l) : kind(tuple_type)
  { tuple=l.release(); }
  inline type_expr(type_ptr arg, type_ptr result); // for function types

  type_expr(const type_expr& t); // copy constructor
  ~type_expr();

  void set_from(type_ptr p);
  bool specialise(const type_expr& pattern);
private:
  type_expr& operator=(const type_expr& t);
   // assignment is forbidden
};


struct type_node
{ type_expr t;  type_list next;

  type_node(type_ptr head, type_list_ptr tail) : t(),next(tail.release())
  { t.set_from(head); }
  type_node(const type_node& n);
  ~type_node();
};
 

struct func_type
{ type_expr arg_type, result_type;

  func_type(type_ptr a, type_ptr r)
   : arg_type(), result_type()
  { arg_type.set_from(a); result_type.set_from(r); }
  func_type(const func_type& f)
   : arg_type(f.arg_type),result_type(f.result_type) {}
};
typedef func_type* func_type_p;
typedef std::auto_ptr<func_type> func_type_ptr;

struct value_base
{ value_base()  {};
  virtual ~value_base()  {};
  virtual void print(std::ostream& out) const =0;
  virtual value_base* clone() const =0;
  static const char* name(); // just a model; this instance remains undefined
private: //copying and assignment forbidden
  value_base& operator=(const value_base& x);
  value_base(const value_base& x);
};

typedef value_base* value;
typedef std::auto_ptr<value_base> owned_value;
typedef std::tr1::shared_ptr<value_base> shared_value;

struct row_value : public value_base
{ std::vector<shared_value> val;

  explicit row_value(size_t n) : val(n) {} // start with |n| empty pointers
  void print(std::ostream& out) const;
  size_t length() const { return val.size(); }
  row_value* clone() const { return new row_value(*this); }
  static const char* name() { return "row value"; }
protected:
  row_value(const row_value& v) : val(v.val) {}
};

typedef std::auto_ptr<row_value> row_ptr;
typedef std::tr1::shared_ptr<row_value> shared_row;

struct tuple_value : public row_value
{ tuple_value(size_t n) : row_value(n) {}
  tuple_value* clone() const { return new tuple_value(*this); }
  void print(std::ostream& out) const;
  static const char* name() { return "tuple value"; }
private:
  tuple_value(const tuple_value& v) : row_value(v){}
 // copy constructor; used by |clone|
};

typedef std::auto_ptr<tuple_value> tuple_ptr;
typedef std::tr1::shared_ptr<tuple_value> shared_tuple;

class context;
typedef std::tr1::shared_ptr<context> context_ptr;
class context
{ const context_ptr next;
  std::vector<shared_value> frame;
  context(const context&); // contexts should not be copied, just shared
public:
  context(const context_ptr& n,
          const std::vector<shared_value>& f) : next(n), frame(f) {}
  shared_value& elem(size_t i,size_t j);
};

struct expression_base
{ enum level { no_value, single_value, multi_value };

  expression_base()  {}
  virtual ~expression_base()  {}
  virtual void evaluate(level l) const =0;
  virtual void print(std::ostream& out) const =0;

  void void_eval() const { evaluate(no_value); }
  void eval() const { evaluate(single_value); }
  void multi_eval() const { evaluate(multi_value); }
};

typedef expression_base* expression;
typedef std::auto_ptr<expression_base> expression_ptr;
typedef std::tr1::shared_ptr<expression_base> shared_expression;

struct conversion_info
{ typedef void (*conv_f)();
  conv_f convert;
  std::string name;
  conversion_info(const char* s,conv_f c) : convert(c),name(s) {}
};

class conversion : public expression_base
{ const conversion_info& type;
  expression exp;
public:
  conversion(const conversion_info& t,expression_ptr e)
   :type(t),exp(e.release()) {}
  virtual ~conversion(){ delete exp; }
  virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const;
};

struct conversion_record : public conversion_info
{ const type_expr* from,* to; // non-owned pointers
  conversion_record (const type_expr& from_type,
                     const type_expr& to_type,
                     const char* s, conv_f c)
   : conversion_info(s,c), from(&from_type),to(&to_type) {}
};

class program_error : public std::exception
{ std::string message;
public:
  explicit program_error(const std::string& s) : message(s) {}
  virtual ~program_error() throw() {} // obligatory definition
  const char* what() const throw() { return message.c_str(); }
};

struct expr_error : public program_error
{ expr offender; // the subexpression causing a problem

  expr_error (const expr& e,const std::string& s) throw()
    : program_error(s),offender(e) {}
};

struct type_error : public expr_error
{ type_p actual; 
  type_p required; // the types that conflicted

  type_error (const expr& e, type_ptr a, type_ptr r) throw() 
    : expr_error(e,"Type error") 
      ,actual(a.release()),required(r.release()) {}
  type_error(const type_error& e);
  ~type_error() throw() { delete actual; delete required; }
};


extern const char* prim_names[];

extern const type_expr unknown_type; // \.{*}
extern type_expr void_type; // \.{()}
extern type_expr int_type; // \.{int}
extern type_expr bool_type; // \.{bool}
extern const type_expr row_of_type; // \.{[*]}
extern const type_expr gen_func_type; 

extern std::vector<shared_value> execution_stack;


type_ptr copy(const type_expr& t);

type_list_ptr make_type_singleton(type_ptr t);
type_list_ptr make_type_list(type_ptr t,type_list_ptr l);
size_t length(type_list l);

std::ostream& operator<<(std::ostream& out, const type_expr& t);

bool operator== (const type_expr& x,const type_expr& y);
inline bool operator!= (const type_expr& x,const type_expr& y)
{ return !(x==y); }

type_ptr make_undetermined_type();
type_ptr make_prim_type(primitive_tag p);
type_ptr make_row_type(type_ptr c);
type_ptr make_tuple_type (type_list_ptr l);
type_ptr make_function_type (type_ptr a, type_ptr r);

type_ptr make_type(const char* s);

type_ptr unknown_tuple(size_t n);

std::ostream& operator<< (std::ostream& out, const value_base& v);

void push_tuple_components();
void wrap_tuple(size_t n);

inline std::ostream& operator<< (std::ostream& out, const expression_base& e)
{ e.print(out); return out; }

bool coerce(const type_expr& from_type, const type_expr& to_type,
            expression_ptr& e);

void coercion(const type_expr& from,
              const type_expr& to,
              const char* s, conversion_info::conv_f f);

conversion_record* row_coercion(const type_expr& final_type,
                                      type_expr& component_type);

unsigned int is_close (const type_expr& x, const type_expr& y);


type_expr::type_expr(type_ptr arg, type_ptr result)
   : kind(function_type) { func=new func_type(arg,result); }

template <typename D> // |D| is a type derived from |value_base|
 D* force(value v) throw(std::logic_error)
{ D* p=dynamic_cast<D*>(v);
  if (p==NULL) throw
    std::logic_error(std::string("forced value is no ")+D::name());
  return p;
}

template<typename D> // |D| is a type derived from |value_base|
  inline void push_value(std::auto_ptr<D> v)
  { execution_stack.push_back(std::tr1::shared_ptr<D>(v)); }

template<typename D> // |D| is a type derived from |value_base|
  inline void push_value(std::tr1::shared_ptr<D> v)
  { execution_stack.push_back(v); }

inline void push_value(value_base* p) { push_value(owned_value(p)); }

inline shared_value pop_value()
{ shared_value arg=execution_stack.back();
  execution_stack.pop_back();
  return arg;
}

template <typename D> // |D| is a type derived from |value_base|
 inline std::tr1::shared_ptr<D> get() throw(std::logic_error)
{ std::tr1::shared_ptr<D> p=std::tr1::dynamic_pointer_cast<D>(pop_value());
  if (p.get()==NULL)
    throw std::logic_error(std::string("Argument is no ")+D::name());
  return p;
}

inline void uniquify(shared_value& v)
{ if (not v.unique()) v=shared_value(v->clone()); }

template <typename D> // |D| is a type derived from |value_base|
 std::tr1::shared_ptr<D> get_own() throw(std::logic_error)
{ uniquify(execution_stack.back());
    return get<D>();
}

} }
#endif

